#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# Main window of the application

""" main window application dialog """

import json
import os
import sys

from .qtapi import qt_api

try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt

from taurus.qt.qtgui.util.ui import UILoadable


from taurus.qt.qtgui.base import TaurusBaseWidget
from .ServerState import ServerState, Checker

from .Detectors import Detectors
from .Preferences import Preferences
from .Descriptions import Descriptions
from .Data import Data
from .Storage import Storage
from .CommandThread import CommandThread
from .MessageBox import MessageBox

from . import __version__
from . import qrc

import logging
import gc
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


@UILoadable(with_ui='ui')
class Selector(Qt.QDialog, TaurusBaseWidget):
    """ main window application dialog """

    #: (:class:`taurus.qt.Qt.pyqtSignal`) doorName  signal
    doorName = Qt.pyqtSignal(str)

    #: (:class:`taurus.qt.Qt.pyqtSignal`)
    # experimentConfigurationChanged signal
    experimentConfigurationChanged = Qt.pyqtSignal(dict)

    def __init__(self, server=None, door=None,
                 standalone=False, umode=None,
                 setdefault=False,
                 switch=True,
                 organization='DESY',
                 application='NXS Component Selector',
                 parent=None):
        """ constructor

        :param server: selector server name
        :type server: :obj:`str`
        :param door: door device name
        :type door: :obj:`str`
        :param standalone: application run without macrogui
        :type standalone: :obj:`bool`
        :param umode: user mode, i.e. simple, user, advanced, special, \
                        expert, administrator
        :type umode: :obj:`str`
        :param setdefault: set default
        :type setdefault: :obj:`bool`
        :param organization: organization name
        :type organization: :obj:`str`
        :param application: application name
        :type application: :obj:`str`
        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        Qt.QWidget.__init__(self, parent)
        TaurusBaseWidget.__init__(self, 'NXSExpDescriptionEditor')
        self.setWindowFlags(Qt.Qt.Window)
        self.loadUi()
        self.debug("SELECTOR load")
        if umode != 'administrator':
            self.ui.tabWidget.removeTab(4)
        logger.debug("PARAMETERS: %s %s", server, parent)
        #: (:obj:`str`) organization name
        self.__organization = organization
        #: (:obj:`str`) application name
        self.__application = application
        #: (:obj:`str`) door name
        self.__door = door
        #: (:obj:`str`) skip configuration reset
        self.__skipconfig = True
        #: (:obj:`bool`) application executed without macrogui
        self.__standalone = standalone

        #: (:obj:`bool`) progressbar is running
        self.__progressFlag = False
        #: (:obj:`bool`) door has to be updated
        self.__doortoupdateFlag = False
        #: (:obj:`bool`) selector server has to be updated
        self.__servertoupdateFlag = False

        #: (:obj:`str`) selector server
        self.__model = None
        #: (:class:`nxsselector.ServerState.ServerState`) server state
        self.state = None

        #: (:obj:`bool`) set default configuration
        self.__setdefault = setdefault
        #: (:obj:`bool`) switch MntGrp into ActiveMntGrp
        self.__switch = switch
        #: (:obj:`str`) user mode
        self.__umode = umode
        #: (:obj:`bool`)  expert mode on
        self.expert = True
        #: (:obj:`bool`)  user mode on
        self.user = False
        #: (:obj:`bool`)  simple mode on
        self.simple = False
        #: (:obj:`bool`)  negative hidden mode on
        self.hidden = False

        #: (:obj:`bool`)  if QSettings loaded
        self.__settingsloaded = False
        if self.__umode:
            self.__setmode(self.__umode)
        #: (:obj:`str`)  configuration file name
        self.cnfFile = ''
        #: (:class:`taurus.qt.Qt.QProgressDialog`) progress bar
        self.__progress = None
        #: (:obj:`list` <:class:`nxsselector.CommandThread.CommandThread`>) \
        #:     command thread
        self.__commandthread = None

        self.__resetServer(server)
        #: (:obj:`int`) user data tab number
        self.__datatab = 2

    def __setmode(self, umode):
        """ sets user mode

        :param umode: user mode, i.e. simple, user, advanced, expert, \
                        administrator
        :type umode: :obj:`str`
        """
        if umode == 'expert':
            self.expert = True
            self.user = False
            self.simple = False
            self.hidden = False
        elif umode == 'advanced':
            self.expert = False
            self.user = False
            self.simple = False
            self.hidden = False
        elif umode == 'user':
            self.expert = False
            self.user = True
            self.simple = False
            self.hidden = False
        elif umode == 'simple':
            self.expert = False
            self.user = True
            self.simple = True
            self.hidden = True
        elif umode == 'special':
            self.expert = True
            self.user = False
            self.simple = False
            self.hidden = True

    def settings(self):
        """ sets configuration
        """
        logger.debug("settings")

        if self.__progress:
            self.__progress.reset()
            self.__progress.hide()
        if self.__commandthread and self.__commandthread.error:
            text = MessageBox.getText(
                "Problems in updating Channels",
                self.__commandthread.error)
            MessageBox.warning(
                self, "NXSelector: Error in Setting Selector Server", text,
                "%s" % str(self.__commandthread.error))
            self.__commandthread.error = None

        settings = Qt.QSettings(self.__organization, self.__application, self)
        if not self.__umode or self.__umode == 'None':
            self.__umode = str(settings.value("Selector/DefaultMode",
                                              self.__umode))
            if not self.__umode or self.__umode == 'None':
                self.__umode = 'expert'
        self.__setmode(self.__umode)
        self.userView = settings.value('Preferences/UserView',
                                       'CheckBoxes Prop')
        self.rowMax = int(settings.value('Preferences/RowMax', 16))
        if not self.rowMax:
            self.rowMax = 16

        self.fontSize = int(self.font().pointSize())
        self.fontSize = int(
            settings.value('Preferences/FontSize', self.fontSize))
        self.displayStatus = int(
            settings.value('Preferences/DisplayStatus', 2))
        self.cnfFile = str(settings.value("Selector/CnfFile", "./"))

        self.__addButtonBoxes()

        # user interface
        self.preferences = Preferences(self.ui, self.state)
        if self.userView not in self.preferences.views:
            self.userView = 'CheckBoxes Dis'
        self.storage = Storage(self.ui, self.state, self.simple)
        self.state.synchthread.scanidchanged.connect(
            self.storage.updateScanID, Qt.Qt.DirectConnection)
        self.state.synchthread.mgconfchanged.connect(
            self.checkDirty, Qt.Qt.DirectConnection)
        self.state.serverChanged.connect(
            self.resetServer, Qt.Qt.DirectConnection)
        self.state.synchthread.restart()

        self.detectors = Detectors(
            self.ui, self.state,
            self.preferences.views[self.userView],
            self.rowMax,
            int(self.hidden) + 2 * int(self.user),
            self.fontSize)

        self.preferences.mgroups = settings.value(
            'Preferences/Groups', '{}')
        self.preferences.frames = settings.value(
            'Preferences/Frames', '[]')
        self.preferences.mgroupshelp = settings.value(
            'Preferences/GroupsHints',
            self.preferences.mgroupshelp)
        self.preferences.frameshelp = settings.value(
            'Preferences/FramesHints',
            self.preferences.frameshelp)

        self.preferences.addHint(
            self.preferences.mgroups,
            self.preferences.mgroupshelp)
        self.preferences.addHint(
            self.preferences.frames,
            self.preferences.frameshelp)

        self.preferences.layoutFile = str(
            settings.value("Preferences/LayoutFile", "./"))

        self.detectors.mgroups = str(self.preferences.mgroups)
        self.detectors.frames = str(self.preferences.frames)
        self.descriptions = Descriptions(
            self.ui, self.state,
            self.preferences.views["CentralCheckBoxes"],
            self.preferences.views["CheckBoxes Dis (U)"],
            self.rowMax,
            self.fontSize)

        self.data = Data(self.ui, self.state, self.simple or self.user)

        self.tabs = [self.detectors, self.descriptions, self.data,
                     self.storage, self.preferences]

        self.waitForThread()
        try:
            if self.__switch:
                self.resetClickAll()
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText(
                "Problems in Switching MntGrp")
            MessageBox.warning(
                self, "NXSelector: Error in Switching MntGrp",
                text, str(value))
        self.createGUI()
        self.__tabChanged(0)

        sg = settings.value("Selector/Geometry")
        if sg:
            self.restoreGeometry(sg)

#        self.resize(0, 0)

        self.title = 'NeXus Component Selector'
        self.__dirty = True
        self.setDirty()
        self.__progress = None
        if self.__servertoupdateFlag:
            self.setModel(self.__model)
        if self.__doortoupdateFlag:
            self.updateDoorName(self.__door)

        self.__settingsloaded = True
        self.__skipconfig = False
        logger.debug("settings END")

    def createGUI(self):
        """ creates GUI for the main window
        """

        self.__setButtonBoxes()
        self.__hideWidgets()
        for tab in self.tabs:
            tab.reset()

        self.__setWidgetValues()
        self.__connectSignals()
        self.__addTips()

    def __addButtonBoxes(self):
        """adds button boxes into the main buttonbox"""
        if not self.__standalone:
            self.ui.mntServerLineEdit.hide()
            self.ui.mntServerLabel.hide()
            self.ui.buttonBox.setStandardButtons(
                Qt.QDialogButtonBox.Reset | Qt.QDialogButtonBox.Apply)
        else:
            self.ui.buttonBox.setStandardButtons(
                Qt.QDialogButtonBox.Reset | Qt.QDialogButtonBox.Apply
                | Qt.QDialogButtonBox.Close)
        self.ui.buttonBox.setSizePolicy(Qt.QSizePolicy.Expanding,
                                        Qt.QSizePolicy.Fixed)

        self.ui.clearAllPushButton = self.ui.buttonBox.addButton(
            Qt.QDialogButtonBox.RestoreDefaults)
        self.ui.clearAllPushButton.setText("ClearAll")
        self.ui.unlockPushButton = self.ui.buttonBox.addButton(
            Qt.QDialogButtonBox.RestoreDefaults)
        self.ui.unlockPushButton.setText("Unlock")
        self.ui.unlockPushButton.hide()

        self.ui.loadProfilePushButton = self.ui.buttonBox.addButton(
            Qt.QDialogButtonBox.Open)
        self.ui.loadProfilePushButton.setText("Load")

        self.ui.saveProfilePushButton = self.ui.buttonBox.addButton(
            Qt.QDialogButtonBox.Save)

        self.ui.groupsPushButton = self.ui.buttonBox.addButton(
            Qt.QDialogButtonBox.Retry)
        self.ui.groupsPushButton.setText("Others")

        self.ui.statusLabel = self.ui.buttonBox.addButton(
            "", Qt.QDialogButtonBox.ActionRole)
        self.ui.statusLabel.setEnabled(False)
        self.ui.buttonBox.setCenterButtons(True)

    def __setButtonBoxes(self):
        """sets button boxes in the main buttonbox"""

        flayout = Qt.QHBoxLayout(self.ui.timerButtonFrame)
        flayout.setContentsMargins(0, 0, 0, 0)
        self.ui.timerAddPushButton = Qt.QPushButton("+")
        self.ui.timerAddPushButton.setMaximumWidth(30)
        flayout.addWidget(self.ui.timerAddPushButton)
        self.ui.timerDelPushButton = Qt.QPushButton("-")
        self.ui.timerDelPushButton.setMaximumWidth(30)
        flayout.addWidget(self.ui.timerDelPushButton)
        self.storage.connectTimerButtons()

    def __setWidgetValues(self):
        """sets widget values"""

        self.ui.layoutButtonBox.button(
            Qt.QDialogButtonBox.Open).setText("Load")
        self.ui.layoutButtonBox.button(
            Qt.QDialogButtonBox.Save).setText("Save")

        layout = self.ui.layoutButtonBox.layout()
        for i in range(layout.count()):
            spacer = layout.itemAt(i)
            if isinstance(spacer, Qt.QSpacerItem):
                spacer.changeSize(
                    0, 0, Qt.QSizePolicy.Minimum)

        self.ui.defaultPushButton = Qt.QPushButton()
#        self.ui.defaultPushButton.resize(0,0)
        self.ui.defaultPushButton.setDefault(True)
#        self.ui.defaultPushButton.setAutoDefault(True)
        layout.addWidget(self.ui.defaultPushButton)
        self.ui.defaultPushButton.hide()

        self.ui.statusLabel.setAutoFillBackground(True)
        self.ui.statusLabel.setSizePolicy(Qt.QSizePolicy.Expanding,
                                          Qt.QSizePolicy.Fixed)

        self.ui.statusLabel.setStyleSheet(
            "background-color:white;border-style: outset; "
            "border-width: 1px; border-color: gray; "
            "color:#208020;font:bold;font-size: 14pt;")
        self.ui.mntGrpComboBox.setInsertPolicy(
            self.ui.mntGrpComboBox.InsertAtTop)

        self.storage.updateMntGrpComboBox()

        cid = self.ui.viewComboBox.findText(str(self.userView))
        if cid >= 0:
            self.ui.viewComboBox.setCurrentIndex(cid)
        self.ui.rowMaxSpinBox.setValue(self.rowMax)
        self.ui.fontSizeSpinBox.setValue(self.fontSize)
        self.ui.statusCheckBox.setChecked(self.displayStatus != 0)

    def __hideWidgets(self):
        """ hides widgets according to set user mode
        """
        if not self.expert:
            self.ui.groupGroupBox.hide()
            self.ui.frameGroupBox.hide()
            self.ui.dynFrame.hide()
            self.ui.devConfigPushButton.hide()
            self.ui.devSettingsLineEdit.setEnabled(False)
            self.ui.devWriterLineEdit.setEnabled(False)
            self.ui.devConfigLineEdit.setEnabled(False)
            self.ui.mntServerLineEdit.setEnabled(False)
        if self.user:
            self.ui.componentGroupBox.hide()
            self.ui.mntGrpToolButton.hide()
            self.ui.viewServerFrame.hide()
            self.ui.viewGroupBox.hide()
            self.ui.tabWidget.removeTab(3)
        if self.simple:
            self.ui.mntGrpComboBox.setEnabled(False)
            self.ui.orderToolButton.hide()
            self.ui.timerAddPushButton.hide()
            self.ui.timerDelPushButton.hide()
            self.ui.mntTimerComboBox.setEnabled(False)
            self.ui.timerButtonFrame.setEnabled(False)
            self.ui.descriptions.hide()
            self.ui.tabWidget.removeTab(3)
            self.ui.tabWidget.removeTab(1)
            self.__datatab -= 1
            self.ui.tabWidget.setCurrentIndex(self.__datatab)

    def __connectSignals(self):
        """ connects all signals
        """
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Apply).pressed.connect(self.__applyClicked)
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Reset).pressed.connect(self.__resetClicked)
        if self.__standalone:
            self.ui.buttonBox.button(
                Qt.QDialogButtonBox.Close).pressed.connect(self.close)
        self.ui.clearAllPushButton.pressed.connect(self.__restore)
        self.ui.unlockPushButton.pressed.connect(self.__unlock)

        self.ui.loadProfilePushButton.pressed.connect(self.cnfLoad)
        self.ui.saveProfilePushButton.pressed.connect(self.cnfSave)

        self.preferences.serverChanged.connect(self.resetServer)
        self.preferences.layoutChanged.connect(self.resetLayout)

        self.ui.viewComboBox.currentIndexChanged.connect(self.resetViews)
        self.ui.rowMaxSpinBox.editingFinished.connect(self.resetRows)
        self.ui.fontSizeSpinBox.editingFinished.connect(self.resetRows)
        self.ui.statusCheckBox.stateChanged.connect(
            self.__displayStatusChanged)

        self.detectors.dirty.connect(self.setDirty)
        self.preferences.dirty.connect(self.setDirty)
        self.descriptions.componentChecked.connect(self.__componentChanged)
        self.data.dirty.connect(self.setDirty)
        self.storage.dirty.connect(self.setDirty)
        self.storage.resetViews.connect(self.resetViews)
        self.storage.resetAll.connect(self.resetAll)
        self.storage.updateGroups.connect(self.updateGroups)
        self.ui.tabWidget.currentChanged.connect(self.__tabChanged)

    def __addTips(self):
        """ adds button tips
        """
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.RestoreDefaults).setToolTip(
            "Deselect all detector components")
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Apply).setToolTip(
            "Send the current profile into the selector server\n" +
            "Update the current measurement group" +
            " and set it into ActiveMntGrp.\n" +
            "Update preselection of description components "
        )
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Reset).setToolTip(
            "Reset local modifications " +
            "by fetching the current profile from the selector server\n" +
            "and updating the profile by the Active Measurement Group.\n" +
            "Update preselection of description components ")
        if self.ui.buttonBox.button(
                Qt.QDialogButtonBox.Close):
            self.ui.buttonBox.button(
                Qt.QDialogButtonBox.Close).setToolTip(
                "Close the Component Selector")
        self.ui.timerAddPushButton.setToolTip("Add a non-master timer")
        self.ui.timerDelPushButton.setToolTip(
            "Remove the last non-master timer ")
        self.ui.loadProfilePushButton.setToolTip(
            "Load a profile with previously selected channels from a file.")
        self.ui.saveProfilePushButton.setToolTip(
            "Save the current profile with the selected channels into a file.")
        self.ui.layoutButtonBox.button(
            Qt.QDialogButtonBox.Open).setToolTip(
            "Load from a file a previously saved detector layout.")
        self.ui.layoutButtonBox.button(
            Qt.QDialogButtonBox.Save).setToolTip(
            "Save into a file the currently detector layout.")
        self.ui.groupsPushButton.setToolTip(
            "Change the available components in the Detectors tab")

    def setModel(self, model):
        """ sets model

        :param model: selector server model
        :type model: :obj:`str`
        """
        if str(model) != str(self.state.server):
            if self.__progress:
                self.__model = model
                self.__servertoupdateFlag = True
            elif self.__servertoupdateFlag:
                self.__servertoupdateFlag = False
                self.ui.devSettingsLineEdit.setText(model)
                self.preferences.changeServer(False)
                logger.debug("change ServerName %s " % model)
        else:
            self.__servertoupdateFlag = False

    def __saveSettings(self):
        """ saves settings
        """
        settings = Qt.QSettings(self.__organization, self.__application, self)
        settings.setValue(
            "Selector/Geometry",
            (self.saveGeometry()))
        settings.setValue(
            "Preferences/UserView",
            (str(self.ui.viewComboBox.currentText())))
        settings.setValue(
            "Preferences/RowMax",
            (self.ui.rowMaxSpinBox.value()))
        settings.setValue(
            "Preferences/FontSize",
            (self.ui.fontSizeSpinBox.value()))
        settings.setValue(
            "Preferences/DisplayStatus",
            (2 if self.ui.statusCheckBox.isChecked() else 0))
        settings.setValue(
            "Preferences/Groups",
            (str(self.preferences.mgroups)))
        settings.setValue(
            "Preferences/Frames",
            (str(self.preferences.frames)))
        settings.setValue(
            "Preferences/FramesHints",
            (self.preferences.frameshelp))
        settings.setValue(
            "Preferences/GroupsHints",
            (self.preferences.mgroupshelp))
        settings.setValue(
            "Selector/CnfFile", (self.cnfFile))
        settings.setValue(
            "Preferences/LayoutFile",
            (self.preferences.layoutFile))
        if self.__setdefault:
            settings.setValue(
                "Selector/DefaultMode",
                (self.__umode))

    def keyPressEvent(self, event):
        """ adds saving settings on escape key

        :param event: key event
        :type event: :class:`taurus.qt.Qt.QEvent`
        """
        if hasattr(event, "key") and event.key() == Qt.Qt.Key_Escape:
            logger.debug("escape key event")
            if self.__settingsloaded:
                self.__saveSettings()
        Qt.QDialog.keyPressEvent(self, event)

    def closeEvent(self, event):
        """ adds saving settings on close event

        :param event: close event
        :type event: :class:`taurus.qt.Qt.QEvent`
        """
        logger.debug("close event")
        if self.__settingsloaded:
            self.__saveSettings()
        Qt.QDialog.closeEvent(self, event)
        logger.debug("close event ended")

    @Qt.pyqtSlot()
    def checkDirty(self):
        if not self.state.isMntGrpChanged():
            self.setDirty()

    def __resetServer(self, server):
        """ resets server state variables

        :param server: server name
        :type server: :obj:`str`
        """
        if self.state:
            with Qt.QMutexLocker(self.state.mutex):
                self.state.synchthread.running = False
            if hasattr(self, "storage"):
                self.state.synchthread.mgconfchanged.disconnect(
                    self.checkDirty)
                self.state.synchthread.scanidchanged.disconnect(
                    self.storage.updateScanID)
                self.state.serverChanged.disconnect(self.resetServer)
            self.state.synchthread.wait()
        try:
            self.state = ServerState(server)
            if self.__door:
                self.state.storeData("door", self.__door)
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText(
                "Problems in resetting Server")
            MessageBox.warning(
                self, "NXSelector: Error in Setting Selector Server",
                text, str(value))
            self.state = ServerState("")
            self.state.setServer()
            if self.__door:
                self.state.storeData("door", self.__door)
        if hasattr(self, "storage"):
            self.state.synchthread.scanidchanged.connect(
                self.storage.updateScanID, Qt.Qt.DirectConnection)
            self.state.synchthread.mgconfchanged.connect(
                self.checkDirty, Qt.Qt.DirectConnection)
            self.state.serverChanged.connect(
                self.resetServer, Qt.Qt.DirectConnection)
            self.state.synchthread.restart()
        if self.__switch:
            cmds = ["fetchSettings"]
        else:
            cmds = ["updateControllers",
                    "fetchSettings"]
        self.runProgress(cmds, "settings")

    def __resetStateThread(self):
        """ resets server state variables
        """
        if self.state:
            with Qt.QMutexLocker(self.state.mutex):
                self.state.synchthread.running = False
            if hasattr(self, "storage"):
                self.state.synchthread.scanidchanged.disconnect(
                    self.storage.updateScanID)
                self.state.synchthread.mgconfchanged.disconnect(
                    self.checkDirty)
            self.state.synchthread.wait()
            if hasattr(self, "storage"):
                self.state.synchthread.scanidchanged.connect(
                    self.storage.updateScanID, Qt.Qt.DirectConnection)
                self.state.synchthread.mgconfchanged.connect(
                    self.checkDirty, Qt.Qt.DirectConnection)
                self.state.synchthread.restart()

    @Qt.pyqtSlot()
    def __componentChanged(self):
        """ updates detector view
        """
        self.setDirty()
        self.detectors.updateViews()

    @Qt.pyqtSlot()
    def setDirty(self, flag=True):
        """ sets dirty flag

        :param flag: dirty flag
        :type flag: :obj:`bool`
        """
        self.__dirty = flag
        self.ui.statusLabel.hide()
        self.ui.buttonBox.button(Qt.QDialogButtonBox.Reset).setEnabled(True)
        self.ui.buttonBox.button(Qt.QDialogButtonBox.Apply).setEnabled(True)
#        self.ui.applyPushButton.setEnabled(True)
#        self.ui.resetPushButton.setEnabled(True)
        self.ui.buttonHorizontalLayout = self.ui.buttonBox.layout()
        spacer = None
        if self.displayStatus:
            self.ui.buttonBox.setCenterButtons(True)
            for i in range(self.ui.buttonHorizontalLayout.count()):
                spacer = self.ui.buttonHorizontalLayout.itemAt(i)
                if isinstance(spacer, Qt.QSpacerItem):
                    spacer.changeSize(
                        0, 0, Qt.QSizePolicy.Minimum)

            if flag:
                self.ui.statusLabel.setStyleSheet(
                    "background-color: white;border-style: outset; "
                    "border-width: 1px; border-color: gray; "
                    "color:#A02020;font:bold;font-size: 14pt;")
                self.ui.statusLabel.setText('NOT APPLIED')
                self.setWindowTitle(
                    '%s (%s mode) * ' % (self.title, self.__umode))
            else:
                self.setWindowTitle(
                    '%s (%s mode)' % (self.title, self.__umode))
                self.ui.statusLabel.setStyleSheet(
                    "background-color: white;border-style: outset; "
                    "border-width: 1px; border-color: gray; "
                    "color:#206020;font:bold;font-size: 14pt;")
                self.ui.statusLabel.setText('APPLIED')
            self.ui.statusLabel.show()
        else:
            self.ui.buttonBox.setCenterButtons(False)
            for i in range(self.ui.buttonHorizontalLayout.count()):
                spacer = self.ui.buttonHorizontalLayout.itemAt(i)
                if isinstance(spacer, Qt.QSpacerItem):
                    spacer.changeSize(
                        40, 20, Qt.QSizePolicy.Expanding)

            if flag:
                self.setWindowTitle(
                    '%s (%s mode) * ' % (self.title, self.__umode))
            else:
                self.setWindowTitle(
                    '%s (%s mode)' % (self.title, self.__umode))
                self.ui.buttonBox.button(
                    Qt.QDialogButtonBox.Reset).setEnabled(False)
                self.ui.buttonBox.button(
                    Qt.QDialogButtonBox.Apply).setEnabled(False)
#                self.ui.applyPushButton.setEnabled(False)
#                self.ui.resetPushButton.setEnabled(False)
            self.ui.statusLabel.hide()

    @Qt.pyqtSlot()
    def resetServer(self):
        """ resets server settings
        """
        logger.debug("reset server")
        self.state.setServer()
        self.__resetStateThread()
        self._resetAll()
        logger.debug("reset server ended")

    @Qt.pyqtSlot(str, str)
    def resetLayout(self, frames, groups):
        """ resets application layout
        """
        logger.debug("reset layout")
        self.detectors.frames = str(frames)
        self.detectors.mgroups = str(groups)
        self.resetViews()
        logger.debug("reset layout ended")

    @Qt.pyqtSlot()
    def resetRows(self):
        """ resets a maximum number of rows in element columns
        """
        logger.debug("reset rows")
        rowMax = self.ui.rowMaxSpinBox.value()
        fontSize = self.ui.fontSizeSpinBox.value()
        for tab in self.tabs:
            if hasattr(tab, "rowMax"):
                tab.rowMax = rowMax
            if hasattr(tab, "fontSize"):
                tab.fontSize = fontSize
        self.resetViews()
        logger.debug("reset rows ended")

    @Qt.pyqtSlot()
    def updateGroups(self):
        """ updates elements groups
        """
        self.state.storeGroups()
        self._resetAll()

    @Qt.pyqtSlot()
    def resetViews(self):
        """ resets all tab views
        """
        logger.debug("reset view")
        self.state.ddsdirty = True
        for tab in self.tabs:
            if not isinstance(tab, Descriptions):
                tab.userView = self.preferences.views[
                    str(self.ui.viewComboBox.currentText())]
            tab.reset()
        logger.debug("reset view end")

    def reset(self):
        """ fetches configuration and resets all tab views
        """
        logger.debug("reset selector")
        try:
            self.state.fetchSettings()
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText("Problems in resetting Server")
            MessageBox.warning(
                self, "NXSelector: Error in Resetting Selector Server",
                text, str(value))
        for tab in self.tabs:
            tab.reset()
        logger.debug("reset selector ended")

    def closeReset(self):
        """ close reset method for progressbar
        """
        status = True
        logger.debug("closing Progress")
        if self.__progress:
            self.__progress.reset()
        if self.__commandthread.error:
            text = MessageBox.getText(
                "Problems in updating Channels",
                self.__commandthread.error)
            MessageBox.warning(
                self,
                "NXSelector: Error in updating Selector Server Channels",
                text,
                "%s" % str(self.__commandthread.error))
            self.__commandthread.error = None
            status = False
        self.reset()
        self.storage.updateMntGrpComboBox()
        self.setDirty(True)
        if self.__progress:
            self.__progress.setParent(None)
            self.__progress = None
        if self.__servertoupdateFlag:
            self.setModel(self.__model)
        if self.__doortoupdateFlag:
            self.updateDoorName(self.__door)
        self.waitForThread()
        gc.collect()
        logger.debug("closing Progress ENDED")
        return status

    def closeResetShowErrors(self):
        self.storage.showErrors()
        self.closeReset()
        self.__skipconfig = False

    def waitForThread(self):
        """ waits for running thread
        """
        logger.debug("waiting for Thread")
        if self.__commandthread:
            self.__commandthread.wait()
        logger.debug("waiting for Thread ENDED")

    def runProgress(self, commands, onclose="closeReset"):
        """ starts progress thread with the given commands

        :param commands: list of commands
        :type commands: :obj:`list` <:obj:`str`>
        :param onclose: close command name
        :type onclose: :obj:`str`
        """
        if self.__progress:
            return
        if self.__commandthread:
            self.__commandthread.setParent(None)
            self.__commandthread = None
        self.__commandthread = CommandThread(self.state, commands, self)
        oncloseaction = getattr(self, onclose)
        self.__commandthread.finished.connect(
            oncloseaction, Qt.Qt.QueuedConnection)
        self.__progress = None
        self.__progress = Qt.QProgressDialog(
            "Updating preselected devices...", "Cancel", 0, 0, self)
        self.__progress.setWindowModality(Qt.Qt.WindowModal)
        self.__progress.setCancelButton(None)
        self.__progress.rejected.connect(
            self.waitForThread, Qt.Qt.QueuedConnection)
        self.__commandthread.start()
        self.__progress.show()

    def resetClickAll(self):
        """ updates the current mntgrp and resets all settings """
        logger.debug("reset ALL")
        self.state.switchMntGrp()
        self._synchDoor()
        self.runProgress(
            [
                "updateControllers",
                "importMntGrp"],
            onclose="closeResetShowErrors")
        logger.debug("reset ENDED")

    def _synchDoor(self):
        """ synchronize the current door name with macroserver
        """
        if not self.__door:
            self.__door = self.state.door
        elif not self.state.isDoorFromMacroServer(self.__door):
            replay = Qt.QMessageBox.question(
                self.ui.preferences,
                "NXSelector: The New Measurement Group"
                "was created for not the current MacroServer.",
                "Would you like to apply change or the current MacroServer"
                " or reset door in the new measurement group? ",
                Qt.QMessageBox.Apply | Qt.QMessageBox.Reset
                | Qt.QMessageBox.Close)
            if replay == Qt.QMessageBox.Apply:
                self.__door = self.state.door
                self.doorName.emit(str(self.state.door))
            elif replay == Qt.QMessageBox.Reset:
                self.updateDoorName(self.__door)
        else:
            self.updateDoorName(self.__door)

    @Qt.pyqtSlot()
    def resetAll(self):
        """ resets all settings and synchronize the current door"""
        self._synchDoor()
        self._resetAll()

    def _resetAll(self):
        """ resets all settings """
        logger.debug("reset ALL")
        self.runProgress(["updateControllers", "importMntGrp"],
                         onclose="closeResetShowErrors")
        logger.debug("reset ENDED")

    def resetDescriptions(self):
        """ resets description selection to the default values"""
        logger.debug("reset Descriptions")
        self.runProgress(["resetDescriptions", "importMntGrp"],
                         onclose="closeResetShowErrors")
        logger.debug("reset Descriptions ENDED")

    def compareConf(self, cnf, ecnf):
        """ copares two dictionaries

        :param cnf: current configuration
        :type cnf: :obj:`dict`
        :param ecnf: external configuration
        :type ecnf: :obj:`dict`
        :returns: if configurations are the same
        :rtype: :obj:`bool`
        """
        if not isinstance(cnf, dict):
            return False
        if not isinstance(ecnf, dict):
            return False

        ekeys = list(set(ecnf.keys()))
        if 'ScanFile' in ekeys:
            if self.state.scanFile != ecnf['ScanFile']:
                return False

        if 'ScanDir' in ekeys:
            if self.state.scanDir != ecnf['ScanDir']:
                return False
        if 'ActiveMntGrp' not in ekeys:
            return False
        amg = ecnf['ActiveMntGrp']
        if amg != cnf['ActiveMntGrp']:
            return False
        if 'MntGrpConfigs' not in ekeys:
            return False
        if not isinstance(cnf['MntGrpConfigs'], dict) or \
           amg not in cnf['MntGrpConfigs']:
            return False
        if not isinstance(ecnf['MntGrpConfigs'], dict) or \
           amg not in ecnf['MntGrpConfigs']:
            return False
        status = Checker().compDict(
            cnf['MntGrpConfigs'][amg], ecnf['MntGrpConfigs'][amg])

        return status

    def resetConfiguration(self, expconf):
        """ resets measurement group configuration

        :param expconf: new measurement group configuration
        :type expconf: :obj:`str`
        """
        logger.debug("reset Configuration")
        if not self.__skipconfig:
            conf = json.loads(self.state.lastMntGrpConfiguration())
            econf = expconf

            if not self.compareConf(conf, econf):
                self.__skipconfig = True
                replay = Qt.QMessageBox.question(
                    self.ui.preferences,
                    "NXSelector: Configuration "
                    "of Measument Group has been changed.",
                    "Would you like to update the changes? ",
                    Qt.QMessageBox.Yes | Qt.QMessageBox.No)
                if replay == Qt.QMessageBox.Yes:
                    self._resetAll()
        logger.debug("reset Configuration END")

    def updateDoorName(self, door):
        """ updates door name

        :param door: new door name
        :type door: :obj:`str`
        """

        logger.debug("update DoorName")
        if self.__progress:
            self.__door = door
            self.__doortoupdateFlag = True
        elif str(door) != str(self.state.door):
            self.__doortoupdateFlag = False
            self.ui.mntServerLineEdit.setText(door)
            self.storage.apply()
            logger.debug("change DoorName %s " % door)

        logger.debug("update DoorName END")

    @Qt.pyqtSlot()
    def __resetClicked(self):
        """ sets reset button focus and resets all settings"""
        self.ui.buttonBox.button(Qt.QDialogButtonBox.Reset).setFocus()
        self.resetClickAll()

    @Qt.pyqtSlot()
    def __restore(self):
        """ unselected all selected elements for detectors
            or reset description to defualt values
        """
        index = self.ui.tabWidget.currentIndex()
        if index == 0:
            self.__clearAllClicked()
        elif index == 1:
            self.resetDescriptions()

    @Qt.pyqtSlot()
    def __unlock(self):
        """ unlock inaccessable descriptive components
        """
        for ds in self.state.idsgroup.keys():
            if self.state.idsgroup[ds] is None:
                self.state.idsgroup[ds] = False
        for cp in self.state.acpgroup.keys():
            if self.state.acpgroup[cp] is None:
                self.state.acpgroup[cp] = False
        self.state.ddsdirty = True
        self.resetViews()
        self.setDirty()

    def __clearAllClicked(self):
        """ unselected all selected elements for detectors
        """
        for ds in self.state.dsgroup.keys():
            self.state.dsgroup[ds] = False
        for ds in self.state.cpgroup.keys():
            self.state.cpgroup[ds] = False
        self.state.ddsdirty = True
        self.resetViews()
        self.setDirty()

    @Qt.pyqtSlot()
    def cnfLoad(self):
        """ loads selection settings from a file"""
        try:
            filename = str(Qt.QFileDialog.getOpenFileName(
                self.ui.storage,
                "Load Profile",
                self.cnfFile,
                "JSON files (*.json);;All files (*)"))
            logger.debug("loading configuration from %s" % filename)
            if filename:
                self.cnfFile = filename
                jconf = open(filename).read()

                self.state.setConfiguration(jconf)
                self.apply()
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText("Problems in loading profile")
            MessageBox.warning(
                self,
                "NXSelector: Error in loading Selector Server profile",
                text, str(value))

    @Qt.pyqtSlot()
    def cnfSave(self):
        """ saves selection settings in a file"""
        try:
            filename = str(Qt.QFileDialog.getSaveFileName(
                self.ui.storage,
                "Save Profile",
                self.cnfFile,
                "JSON files (*.json);;All files (*)"))
            logger.debug("saving configuration to %s" % filename)
            if filename:
                if (len(filename) < 4 or filename[-4] != '.') and \
                        not (len(filename) > 5 and filename[-5] == '.'):
                    filename = filename + '.json'

                jconf = self.state.getConfiguration()
                self.cnfFile = filename

                with open(filename, 'w') as myfile:
                    myfile.write(jconf)
                self._resetAll()

        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText("Problems in saving profile")
            MessageBox.warning(
                self,
                "NXSelector: Error in saving Selector Server profile",
                text, str(value))

    @Qt.pyqtSlot()
    def __applyClicked(self):
        """ focuses on apply button and performs apply action"""
        self.ui.buttonBox.button(Qt.QDialogButtonBox.Apply).setFocus()
        self.__skipconfig = True
        self.apply()

    @Qt.pyqtSlot(int)
    def __tabChanged(self, index):
        """  updates button views on tab change

        :param index:  tag index
        :type index: :obj:`int`
        """
        if index == 0:
            self.ui.groupsPushButton.setText("Others")
            self.ui.groupsPushButton.show()
            self.ui.groupsPushButton.setToolTip(
                "Change the available components in the Detectors tab")
            self.ui.clearAllPushButton.setText("ClearAll")
            self.ui.clearAllPushButton.setToolTip(
                "Deselect all detector components")
            self.ui.clearAllPushButton.show()
            self.ui.unlockPushButton.hide()
        elif index == 1:
            self.ui.groupsPushButton.setText("Others")
            self.ui.groupsPushButton.show()
            self.ui.groupsPushButton.setToolTip(
                "Change the available components in the Descriptions tab")
            self.ui.clearAllPushButton.setText("Reset Desc.")
            self.ui.clearAllPushButton.setToolTip(
                "Reset profile and the description components "
                "into the default set")
            self.ui.clearAllPushButton.show()
            self.ui.unlockPushButton.setToolTip(
                "Unlock inaccessible description components")
            self.ui.unlockPushButton.show()
        else:
            self.ui.unlockPushButton.hide()
            self.ui.clearAllPushButton.hide()
            self.ui.groupsPushButton.hide()

    @Qt.pyqtSlot(int)
    def __displayStatusChanged(self, state):
        """  displays status in buttonbox

        :param state:  status state
        :type index: :obj:`int`
        """
        self.displayStatus = state
        self.setDirty(self.__dirty)

    def closeApply(self):
        """ close action for apply command """
        self.storage.showErrors()
        if not self.closeReset():
            self.setDirty(True)
        else:
            self.setDirty(False)
        self.ui.fileScanIDSpinBox.setValue(self.state.scanID)
        self.__skipconfig = False

    def apply(self):
        """ applies seeting on selection server and
            creates a new measurement group"""
        logger.debug("apply")
        try:
            conf = self.state.updateMntGrp()
            self.experimentConfigurationChanged.emit(conf)
            self.runProgress(["updateControllers", "importMntGrp",
                              "createConfiguration"],
                             "closeApply")
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText("Problems in resetting Server")
            self.setDirty(True)
            if str(text).startswith("Exception: User Data not defined ["):
                ctext = str(text).replace(
                    "Exception: User Data not defined ", "").replace(
                    "'", '"').split('\n')[0].strip()
                self.defineMissingKeys(ctext)
            MessageBox.warning(
                self,
                "NXSelector: Error in applying Selector Server settings",
                text, str(value))

        logger.debug("apply END")

    def defineMissingKeys(self, ctext):
        """ adds missing keys to the user data table

        :param ctext: json list of missing user data keys
        :type ctext: :obj:`str`
        """
        if ctext:
            missingkeys = json.loads(ctext)
            for key in missingkeys:
                self.state.datarecord[key] = ""
            self.data.reset()
            self.ui.tabWidget.setCurrentIndex(self.__datatab)


def main():
    """ the main function
    """
    if "GNOME_DESKTOP_SESSION_ID" not in os.environ:
        os.environ["GNOME_DESKTOP_SESSION_ID"] = "qtconfig"
    if os.path.isdir("/usr/lib/kde4/plugins/") and \
       "QT_PLUGIN_PATH" not in os.environ:
        os.environ["QT_PLUGIN_PATH"] = "/usr/lib/kde4/plugins/"
    try:
        from taurus.external.qt import Qt
    except Exception:
        from taurus.qt import Qt
    try:
        from taurus.external.qt import __qt
    except Exception:
        from taurus.qt import __qt
    Qt.QCoreApplication.setAttribute(Qt.Qt.AA_X11InitThreads)
    Qt.QResource.registerResource(
        os.path.join(qrc.__path__[0], "resources.rcc"))

    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    standalone = app is None
    server = None
    # expert = True
    umode = None
    setdefault = False
    door = None
    switch = True

    logger.debug("Using %s" % qt_api)
    if standalone:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()

        parser.add_option(
            "-s", "--server", dest="server",
            help="selector server")
        parser.add_option(
            "-d", "--door", dest="door",
            help="door device name")
        parser.add_option(
            "-t", "--style", dest="style",
            help="Qt style")
        parser.add_option(
            "-y", "--stylesheet", dest="stylesheet",
            help="Qt stylesheet")
        parser.add_option(
            "-m", "--mode", dest="mode",
            help="interface mode, i.e. simple, user, advanced, "
            "special, expert")
        parser.add_option(
            "", "--set-as-default-mode",
            action="store_true",
            default=False,
            dest="setdefault",
            help="set the current mode as default")
        parser.add_option(
            "", "--dont-switch-mntgrp",
            action="store_false",
            default=True,
            dest="switch",
            help="do not switch MntGrp to the ActiveMntGrp")

        app = Application(
            sys.argv,
            cmd_line_parser=parser,
            app_name="NXS Component Selector",
            app_version=__version__,
            org_domain="desy.de",
            org_name="DESY")

        app.setWindowIcon(Qt.QIcon(":/configtools.png"))

        (options, _) = parser.parse_args()
        if options.style:
            app.setStyle(options.style)

        server = options.server
        if options.stylesheet:
            app.setStyleSheet(options.stylesheet)
        elif options.style == "cleanlooks":
            pyqtver = __qt.PYQT_VERSION.split(".")
            if pyqtver and pyqtver[0] == "4":
                # fix for cleanlooks tooltip colors
                app.setStyleSheet(
                    "QToolTip{ color: black; background-color: white; }"
                )

        if options.mode:
            umode = options.mode
        else:
            umode = None
        setdefault = options.setdefault
        door = options.door
        switch = options.switch

    form = Selector(server, standalone=True, umode=umode,
                    setdefault=setdefault,
                    door=door,
                    switch=switch)
    form.show()

    if standalone:
        sys.exit(app.exec_())
    else:
        return form


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# editable list of component groups

"""  editable list dialog """

try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt

from taurus.qt.qtgui.util.ui import UILoadable
# from taurus.qt.qtgui.panel import TaurusModelChooser
# from taurus.qt.qtgui.panel import TaurusModelSelectorTree
# from taurus.core.taurusbasetypes import TaurusElementType
# import taurus

from .Views import OneTableView
from .Element import GElement, CP, DS
from .ElementModel import ElementModel
from .AddDataSourceDlg import AddDataSourceDlg


import logging
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


@UILoadable(with_ui='ui')
class GroupsDlg(Qt.QDialog):
    """  editable other device list dialog
    """

    def __init__(self, parent=None):
        """ constructor

        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        Qt.QDialog.__init__(self, parent)
        self.loadUi()
        #: (:obj:`bool`) dirty flag
        self.dirty = False
        #: (:obj:`dict` <:obj:`str`, :obj:`bool` or `None`>) \
        #:     component selection
        self.components = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`bool` or `None`>) \
        #:     datasource selection
        self.datasources = {}
        #: (:obj:`list` <:class:`nxsselector.Element.Element`>) \
        #:    list of component elements
        self.dcpgroup = []
        #: (:obj:`list` <:class:`nxsselector.Element.Element`>) \
        #:    list of datasource elements
        self.ddsgroup = []
        #: (:class:`nxsselector.ServerState.ServerState`) server state
        self.state = None
        #: (:obj:`str`) group title
        self.title = "Selectable Detector Elements"
        #: (:obj:`dict` <:obj:`str`, :obj:`str``>) \
        #:     datasources to add { name: source }
        self.newdatasources = {}

    def __createViews(self, widget, cpview, dsview):
        """ creates basic views

        :param widget: main detector widget
        :type widget: :class:`taurus.qt.Qt.QWidget`
        :param cpview: component table view
        :type cpview: :class:`taurus.qt.Qt.QTableView`
        :param dsview: datasource table view
        :type dsview: :class:`taurus.qt.Qt.QTableView`
        """
        gridLayout_3 = Qt.QGridLayout(widget)
        gridLayout = Qt.QGridLayout()
        gridLayout.addWidget(cpview, 0, 0, 1, 1)
        gridLayout.addWidget(dsview, 0, 1, 1, 1)
        gridLayout_3.addLayout(gridLayout, 0, 0, 1, 1)

    def createGUI(self):
        """ creates GUI
        """
        self.setWindowTitle(self.title)
        self.ui.dcpTableView = OneTableView(self.ui.detector)
        self.ui.ddsTableView = OneTableView(self.ui.detector)

        self.__createViews(self.ui.detector,
                           self.ui.dcpTableView, self.ui.ddsTableView)

        self.__populateTable(self.ui.dcpTableView, self.dcpgroup, CP,
                             self.components, "Components:")
        self.__populateTable(self.ui.ddsTableView, self.ddsgroup, DS,
                             self.datasources, "DataSources:")

        self.ui.dcpTableView.dirty.connect(self.__dirty)
        self.ui.ddsTableView.dirty.connect(self.__dirty)

        self.ui.closeButtonBox.button(
            Qt.QDialogButtonBox.Close).clicked.connect(self.reject)

        self.ui.createPushButton = self.ui.closeButtonBox.addButton(
            "", Qt.QDialogButtonBox.ActionRole)
        self.ui.createPushButton.setText("Create DataSources ...")
        self.ui.createPushButton.clicked.connect(
            self.__createDataSources)

    @Qt.pyqtSlot()
    def __dirty(self):
        """ sets dirty to True
        """
        self.dirty = True
        self.setWindowTitle("Component Groups *")
        logger.debug("changed")

    @Qt.pyqtSlot()
    def __createDataSources(self):
        """ selects configuration of new datasources
        """
        dform = AddDataSourceDlg(self)
        dform.createGUI()
        if dform.exec_():
            self.newdatasources[dform.name] = dform.source
            self.datasources[dform.name] = True
            datasources = {dform.name: True}
            self.__populateTable(self.ui.ddsTableView, self.ddsgroup, DS,
                                 datasources, "DataSources:")
            self.dirty = True

    def __populateTable(self, view, group, eltype, dct, header):
        """ populates the group table

        :param view: element table view
        :type view: :class:`taurus.qt.Qt.QTableView`
        :param group: list ofg elements
        :type group: :obj:`list` <:class:`nxsselector.Element.Element`>
        :param eltype: element type, i.e datasource=DS (0) or component=CP (1)
        :type eltype: :obj:`int`
        :param dct: element selection dictionary
        :type dct: :obj:`dict` <:obj:`str`, :obj:`bool` or `None`>
        :param header: table header
        :type header: :obj:`str`
        """

        for el in dct.keys():
            group.append(GElement(el, eltype, self.state, dct))
        md = ElementModel(group)
        md.headers = [header]
        md.autoEnable = False
        view.horizontalHeader().setVisible(True)
        view.setModel(md)
        md.dirty.connect(self.__dirty)



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# DynamicTools

""" dynamic widget tools """

import logging
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)

try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt


class DynamicTools(object):
    """  dynamic widget tools """

    @classmethod
    def cleanupObjects(cls, objects, label=None):
        """ cleans up qt objects

        :param objects: qt objects to be cleaned up
        :type objects: :obj:`list` <:class:`taurus.qt.Qt.QObject`>
        :param label: qt objects label
        :type label: :obj:`str`
        """

        while objects:
            la = objects.pop()
            try:
                if hasattr(la, "deleteLater"):
                    la.deleteLater()
                if label:
                    logger.debug("del %s" % label)
            except Exception:
                if label:
                    logger.debug("ERROR del %s" % label)

    @classmethod
    def cleanupWidgets(cls, widgets, label=None):
        """ cleans up qt widgets

        :param widgets: qt widgets to be cleaned up
        :type widgets: :obj:`list` <:class:`taurus.qt.Qt.QWidget`>
        :param label: qt object label
        :type label: :obj:`str`
        """
        while widgets:
            wg = widgets.pop()
            if hasattr(wg, "clearLayout"):
                wg.clearLayout()
            try:
                wg.hide()
                wg.close()
                if hasattr(wg, "deleteLater"):
                    wg.deleteLater()
                if label:
                    logger.debug("del %s" % label)
            except Exception:
                if label:
                    logger.debug("ERROR del %s" % label)

    @classmethod
    def cleanupFrames(cls, frames, label=None):
        """ cleans up qt widgets

        :param frames: qt frames to be cleaned up
        :type frames: :obj:`list` <:class:`taurus.qt.Qt.QFrame`>
        :param label: qt object label
        :type label: :obj:`str`
        """
        while frames:
            fr = frames.pop()
            try:
                fr.hide()
                fr.close()
                if label:
                    logger.debug("del %s" % label)
            except Exception:
                if label:
                    logger.debug("ERROR del %s" % label)

    @classmethod
    def cleanupLayoutWithItems(cls, layout):
        """ cleans up qt layout with its items

        :param layout: qt layout to be cleaned up
        :type layout: :class:`taurus.qt.Qt.QLayout`>
        """
        if layout:
            logger.debug("COUNTS %s" % layout.count())
            while layout.count():
                child = layout.itemAt(0)
                layout.removeItem(child)
                if isinstance(child, Qt.QWidgetItem):
                    w = child.widget()
                    w.hide()
                    w.close()
                    layout.removeWidget(w)
                    if hasattr(w, "deleteLater"):
                        w.deleteLater()
                        logger.debug("WL %s" % type(w))
                    else:
                        logger.debug("WW %s" % type(w))
                    w = None
            Qt.QWidget().setLayout(layout)



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# properties dialog

"""  properties dialog """


try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt

from taurus.qt.qtgui.util.ui import UILoadable

from .LDataDlg import LDataDlg

import logging
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


class PropertiesDlg(Qt.QDialog):
    """  properties dialog """

    def __init__(self, parent=None):
        """ constructor

        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        Qt.QDialog.__init__(self, parent)
        #: (:class:`nxselector.PropertiesWg.PropertiesWg`) properties widget
        self.widget = PropertiesWg(self)
        #: (:obj:`bool`) dirty flag
        self.dirty = False
        #: (:obj:`list` <:obj:`str`>) available names
        self.available_names = None

    def createGUI(self):
        """ creates widget GUI
        """
        self.widget.available_names = self.available_names
        self.widget.createGUI()
        layout = Qt.QHBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.widget.ui.closePushButton.clicked.connect(self.accept)
        self.widget.ui.closePushButton.show()
        self.widget.dirty.connect(self.__setDirty)

    def __setDirty(self):
        """ set dirty flag """
        self.dirty = True


@UILoadable(with_ui='ui')
class PropertiesWg(Qt.QWidget):
    """  properties widget """

    #: (:class:`taurus.qt.Qt.pyqtSignal`) dirty signal
    dirty = Qt.pyqtSignal()

    def __init__(self, parent=None):
        """ constructor

        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        Qt.QWidget.__init__(self, parent)
        self.loadUi(filename='EdListWg.ui')
        #: (:obj:`bool`) simple view flag
        self.simple = False
        #: (:obj:`dict` <:obj:`str`, :obj:`str`>) \
        #:    property (name, path) dictionary
        self.paths = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`str`>) \
        #:    property (name, JSON list with shape) dictionary
        self.shapes = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`bool`>) \
        #:    property (name, link) dictionary
        self.links = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`bool`>) \
        #:    property (name, canfailflag) dictionary
        self.canfailflags = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`str`>) \
        #:    property (name, nexus type) dictionary
        self.types = {}
        #: (:obj:`list` <:obj:`str`>) available names
        self.available_names = None

    def createGUI(self):
        """ creates widget GUI
        """
        self.ui.closePushButton = self.ui.closeButtonBox.button(
            Qt.QDialogButtonBox.Close)
        self.ui.addPushButton = self.ui.addEditRemoveButtonBox.addButton(
            "&Add", Qt.QDialogButtonBox.ActionRole)
        self.ui.editPushButton = self.ui.addEditRemoveButtonBox.addButton(
            "&Edit", Qt.QDialogButtonBox.ActionRole)
        self.ui.removePushButton = self.ui.addEditRemoveButtonBox.addButton(
            "&Remove", Qt.QDialogButtonBox.ActionRole)

        names = self.__names()
        if names:
            item = names[0]
        else:
            item = None
        self.__populateTable(item)
        self.ui.addPushButton.clicked.connect(self.__add)
        self.ui.editPushButton.clicked.connect(self.__edit)
        self.ui.tableWidget.itemDoubleClicked.connect(self.__edit)
        self.ui.removePushButton.clicked.connect(self.__remove)

    def __names(self):
        """ provides property names

        :returns:  property names
        :rtype: :obj:`list` <:obj:`str`> >
        """
        return sorted(set(self.paths.keys()) |
                      set(self.shapes.keys()) |
                      set(self.links.keys()) |
                      set(self.canfailflags.keys()) |
                      set(self.types.keys()))

    def __populateTable(self, selected=None):
        """ populates the group table

        :param selected: selected property
        :type selected: :obj:`str`
        """
        self.ui.tableWidget.clear()
        sitem = None
        self.ui.tableWidget.setSortingEnabled(False)
        names = self.__names()
        self.ui.tableWidget.setRowCount(len(names))
        headers = ["Name", "Type", "Shape", "Link", "CanFail", "Path"]
        self.ui.tableWidget.setColumnCount(len(headers))
        self.ui.tableWidget.setHorizontalHeaderLabels(headers)
        for row, name in enumerate(names):
            enable = True
            if self.available_names is not None and\
                    name not in self.available_names:
                enable = False
            item = Qt.QTableWidgetItem(name)
            if self.available_names is not None:
                if enable is False:
                    flags = item.flags()
                    flags &= ~Qt.Qt.ItemIsEnabled
                    item.setFlags(flags)
            if selected is not None and selected == name:
                sitem = item
            self.ui.tableWidget.setItem(row, 0, item)

            value = str(self.types[name]) \
                if name in self.types.keys() else ''
            self.ui.tableWidget.setItem(row, 1, Qt.QTableWidgetItem(value))
            value = str(self.shapes[name]) \
                if name in self.shapes.keys() else ''
            self.ui.tableWidget.setItem(row, 2, Qt.QTableWidgetItem(value))
            value = str(self.links[name]) \
                if name in self.links.keys() else ''
            self.ui.tableWidget.setItem(row, 3, Qt.QTableWidgetItem(value))
            value = str(self.canfailflags[name]) \
                if name in self.canfailflags.keys() else ''
            self.ui.tableWidget.setItem(row, 4, Qt.QTableWidgetItem(value))
            value = str(self.paths[name]) \
                if name in self.paths.keys() else ''
            self.ui.tableWidget.setItem(row, 5, Qt.QTableWidgetItem(value))
        self.ui.tableWidget.resizeColumnsToContents()
        self.ui.tableWidget.setSelectionBehavior(
            Qt.QAbstractItemView.SelectRows)
        self.ui.tableWidget.setSelectionMode(
            Qt.QAbstractItemView.SingleSelection)
        self.ui.tableWidget.horizontalHeader(
        ).setStretchLastSection(True)
        self.ui.tableWidget.setEditTriggers(
            Qt.QAbstractItemView.NoEditTriggers)
        if sitem is not None:
            sitem.setSelected(True)
            self.ui.tableWidget.setCurrentItem(sitem)

    def __currentName(self):
        """ provides a name of the currently selected channel

        :returns: name of the currently selected channel
        :rtype: :obj:`str`
        """
        name = None
        row = self.ui.tableWidget.currentRow()
        skeys = self.__names()
        if len(skeys) > row and row >= 0:
            name = skeys[row]
        return name

    @classmethod
    def __updateItem(cls, name, value, dct):
        """ updates property items in the given dictionary

        :param name: property name
        :type name: :obj:`str`
        :param value: property value
        :type value: `any`
        :param dct: (name, value) dictionary to be updated
        :type dct: :obj:`dict` < :obj:`str`, `any`>
`
        """
        if not value:
            if name in dct.keys():
                dct.pop(name)
        elif name:
            dct[name] = value

    def __updateTable(self, form):
        """ updates table acdcording to the given form

        :param form: LDataDlg form with properties
        :type form: :class:`nxselector.LDataDlg.LDataDlg`
        """
        if form.label:
            name = form.label

            self.__updateItem(name, form.path, self.paths)
            self.__updateItem(name, form.dtype, self.types)
            self.__updateItem(name, form.shape, self.shapes)

            if form.link is None:
                if name in self.links.keys():
                    self.links.pop(name)
            elif name:
                self.links[name] = form.link

            if form.canfail is None:
                if name in self.canfailflags.keys():
                    self.canfailflags.pop(name)
            elif name:
                self.canfailflags[name] = form.canfail

            self.__populateTable()
            self.dirty.emit()

    @Qt.pyqtSlot()
    def __add(self):
        """ adds a new property """
        dform = LDataDlg(self)
        dform.available_names = self.available_names
        dform.createGUI()
        if dform.exec_():
            self.__updateTable(dform)

    @Qt.pyqtSlot()
    def __edit(self):
        """ change a value of the current property """
        dform = LDataDlg(self)
        name = self.__currentName()
        dform.label = name
        if name in self.types.keys():
            dform.dtype = self.types[name]
        if name in self.shapes.keys():
            dform.shape = self.shapes[name]
        if name in self.links.keys():
            dform.link = self.links[name]
        if name in self.canfailflags.keys():
            dform.canfail = self.canfailflags[name]
        if name in self.paths.keys():
            dform.path = self.paths[name]

        dform.available_names = self.available_names
        dform.createGUI()
        if name:
            dform.ui.labelLineEdit.setEnabled(False)
        if dform.exec_():
            self.__updateTable(dform)

    @Qt.pyqtSlot()
    def __remove(self):
        """ remmoves the current property """
        name = self.__currentName()

        if Qt.QMessageBox.question(
                self, "Removing Data",
                "Would you like  to remove '%s'?" % name,
                Qt.QMessageBox.Yes | Qt.QMessageBox.No,
                Qt.QMessageBox.Yes) == Qt.QMessageBox.No:
            return
        if name in self.types.keys():
            self.types.pop(name)
        if name in self.shapes.keys():
            self.shapes.pop(name)
        if name in self.links.keys():
            self.links.pop(name)
        if name in self.canfailflags.keys():
            self.canfailflags.pop(name)
        if name in self.paths.keys():
            self.paths.pop(name)

        self.dirty.emit()
        self.__populateTable()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# editable data dialog

"""  editable data dialog """

import json
import logging
import sys

try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt

from taurus.qt.qtgui.util.ui import UILoadable

from .MessageBox import MessageBox
from. ElementModel import PROPTEXT


#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)

SYNCHTEXT = PROPTEXT["synchronization"]

if sys.version_info > (3,):
    unicode = str


@UILoadable(with_ui='ui')
class LDataDlg(Qt.QDialog):
    """  editable data dialog """

    def __init__(self, parent=None):
        """ constructor

        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        Qt.QDialog.__init__(self, parent)
        self.loadUi()
        #: (:obj:`str`) device label
        self.label = ''
        #: (:obj:`str`) nexus path
        self.path = ''
        #: (:obj:`str`) shape in JSON list
        self.shape = None
        #: (:obj:`str`) nexus data type
        self.dtype = ''
        #: (:obj:`bool`) if nxdata link should be created
        self.link = None
        #: (:obj:`bool`) if ds switched to CanFail mode
        self.canfail = None
        #: (:obj:`list` <:obj:`str` > ) available variable names
        self.available_names = None
        #: (:obj:`list` <:obj:`str` > ) special variable names
        self.special = ["shape", "data_type", "nexus_path", "canfail", "link"]
        #: (:obj:`dict` <:obj:`str` , :obj:`str` or :obj:`unicode`> ) \
        #:     (name, value) variable dictionary
        self.variables = {}
        #: (:obj:`dict` <:obj:`str` , :class:`taurus.qt.Qt.QLabel`> ) \
        #:     (name, qlabel) variable dictionary
        self.names = {}
        #: (:obj:`dict` <:obj:`str` , :class:`taurus.qt.Qt.QWidget`> ) \
        #:     (name, qwidget) variable dictionary
        self.widgets = {}
        #: (:obj:`list` <:obj:`str` > ) special variable names
        self.hidden = ["synchronization", "synchronizer",
                       "entryname", "serialno"]
        #: (:obj:`list` <:obj:`str` > ) special variable names
        self.extra = []
        #: (:obj:`list` <:obj:`str` > ) synchronizers
        self.synchronizers = []

    @classmethod
    def __linkText(cls, value):
        """ converts link value to string, i.e. True, False or Default

        :param value: link value
        :type value: :obj:`bool` or `None`
        :returns: True, False or Default
        :rtype: :obj:`str`
        """
        if isinstance(value, bool):
            if value is True:
                return "True"
            if value is False:
                return "False"
        return "Default"

    @classmethod
    def __canfailText(cls, value):
        """ converts canfail value to string, i.e. True, Default

        :param value: canfail value
        :type value: :obj:`bool` or `None`
        :returns: True, False or Default
        :rtype: :obj:`str`
        """
        if isinstance(value, bool):
            if value is True:
                return "True"
        return "Default"

    def createGUI(self):
        """ creates GUI
        """
        self.ui.synchronizationComboBox.hide()
        self.ui.synchronizationLabel.hide()
        self.ui.synchronizerComboBox.hide()
        self.ui.synchronizerLabel.hide()
        self.ui.labelLineEdit.setText(str(str(self.label or "")))
        self.ui.pathLineEdit.setText(str(str(self.path or "")))
        if self.shape is None:
            shape = ''
        else:
            shape = self.shape
        self.ui.shapeLineEdit.setText(
            str(str(shape)))
        self.ui.typeLineEdit.setText(str(str(self.dtype or "")))

        cid = self.ui.linkComboBox.findText(
            str(self.__linkText(self.link)))
        if cid < 0:
            cid = 0
        self.ui.linkComboBox.setCurrentIndex(cid)

        cid = self.ui.canfailComboBox.findText(
            str(self.__canfailText(self.canfail)))
        if cid < 0:
            cid = 0
        self.ui.canfailComboBox.setCurrentIndex(cid)

        if self.available_names:
            completer = Qt.QCompleter(self.available_names, self)
            self.ui.labelLineEdit.setCompleter(completer)

        if self.variables:
            self.addGrid()
        self.adjustSize()

    def addGrid(self):
        """ adds from grid
        """
        index = 0
        for nm, val in self.variables.items():
            if not nm.startswith("__"):
                if (nm not in self.extra and nm not in self.hidden) or \
                   (val and nm not in self.extra and nm in self.hidden):
                    self.names[nm] = Qt.QLabel(self.ui.varFrame)
                    self.names[nm].setText(str(str(nm)))
                    self.ui.varGridLayout.addWidget(
                        self.names[nm], index, 0, 1, 1)
                    self.widgets[nm] = Qt.QLineEdit(self.ui.varFrame)
                    if val is not None:
                        self.widgets[nm].setText(str(str(val or "")))
                    self.ui.varGridLayout.addWidget(
                        self.widgets[nm], index, 1, 1, 1)
                    index += 1

    def addVariables(self, variables):
        """ adds  variables

        :param variables: (name, value) variable dictionary
        :type variables: :obj:`dict` <:obj:`str`, :obj:`str` or :obj:`unicode`>
        """
        leftchannels = False
        for sp in self.special:
            if sp in variables.keys():
                leftchannels = True
        if not leftchannels:
            self.ui.channelFrame.hide()
        for vr, val in variables.items():
            if vr not in self.special:
                self.variables[vr] = val

    def accept(self):
        """ updates class variables with the form content
        """
        link = str(self.ui.linkComboBox.currentText())
        if link == "True":
            self.link = True
        elif link == "False":
            self.link = False
        else:
            self.link = None

        canfail = str(self.ui.canfailComboBox.currentText())
        if canfail == "True":
            self.canfail = True
        else:
            self.canfail = None

        self.label = unicode(self.ui.labelLineEdit.text())
        self.path = unicode(self.ui.pathLineEdit.text())
        self.dtype = unicode(self.ui.typeLineEdit.text())
        tshape = unicode(self.ui.shapeLineEdit.text()).replace("None", "null")
        try:
            if not tshape:
                self.shape = None
            else:
                shape = json.loads(tshape)
                if not isinstance(shape, list):
                    raise Exception("shape is not a list")
                self.shape = shape
        except Exception:
            import traceback
            value = traceback.format_exc()
            text = MessageBox.getText("Wrong structure of Shape")
            MessageBox.warning(
                self,
                "NXSelector: Wrong Data",
                text, str(value))
            self.ui.shapeLineEdit.setFocus()
            return

        for nm, wg in self.widgets.items():
            if nm not in self.extra:
                self.variables[nm] = unicode(wg.text()) or None

        if not self.label:
            Qt.QMessageBox.warning(self, "Wrong Data", "Empty data label")
            self.ui.labelLineEdit.setFocus()
            return

        Qt.QDialog.accept(self)


class LExDataDlg(LDataDlg):
    """  extension of editable data dialog """

    def __init__(self, parent=None):
        """ constructor

        :param parent: parent object
        :type parent: :class:`taurus.qt.Qt.QObject`
        """
        LDataDlg.__init__(self, parent)
        #: (:obj:`list` <:obj:`str` > ) special variable names
        self.extra = ["synchronization", "synchronizer"]

    def createGUI(self):
        """ creates GUI
        """
        LDataDlg.createGUI(self)
        self.ui.synchronizationComboBox.show()
        self.ui.synchronizationLabel.show()
        self.ui.synchronizerComboBox.show()
        self.ui.synchronizerLabel.show()

        if self.variables and "synchronization" in self.variables:
            synch = int(self.variables["synchronization"] or 0)
        else:
            synch = 0
        cid = self.ui.synchronizationComboBox.findText(
            str(SYNCHTEXT[synch]))
        if cid < 0:
            cid = 0
        self.ui.synchronizationComboBox.setCurrentIndex(cid)

        if self.variables and "synchronizer" in self.variables:
            val = self.variables["synchronizer"] or "software"
        else:
            val = "software"

        self.ui.synchronizerComboBox.clear()
        self.ui.synchronizerComboBox.addItem("software")
        for sch in self.synchronizers:
            self.ui.synchronizerComboBox.addItem(sch)

        cid = self.ui.synchronizerComboBox.findText(
            str(str(val)))
        if cid < 0:
            cid = 0
        self.ui.synchronizerComboBox.setCurrentIndex(cid)

#        if val is not None:
#            self.ui.synchronizerLineEdit.setText(
#                str(str(val or "")))
        self.adjustSize()

    def accept(self):
        """ updates class variables with the form content
        """

        textsynch = str(self.ui.synchronizationComboBox.currentText())
        synch = SYNCHTEXT.index(textsynch) \
            if textsynch in SYNCHTEXT else None
        if synch or "synchronization" in self.variables:
            self.variables["synchronization"] = synch

        syncher = unicode(self.ui.synchronizerComboBox.currentText())
        if syncher != "software" or "synchronizer" in self.variables:
            self.variables["synchronizer"] = syncher

        LDataDlg.accept(self)



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# preferences tab

""" preferences tab """

import os
import PyTango
import json

from .MessageBox import MessageBox
from .DynamicTools import DynamicTools
try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt

from .Views import (TableView, OneTableView,
                    CheckerLabelViewNN,
                    CheckerView, RadioView, ButtonView,
                    LeftCheckerView, LeftRadioView,
                    CheckerViewNL, RadioViewNL, ButtonViewNL,
                    LeftCheckerViewNL, LeftRadioViewNL,
                    CheckerViewNN, RadioViewNN, ButtonViewNN,
                    LeftCheckerViewNN, LeftRadioViewNN,
                    CheckDisView, RadioDisView,
                    CheckDisViewNL, RadioDisViewNL,
                    CheckDisViewNN, RadioDisViewNN,
                    CheckPropView,
                    CheckPropViewNL,
                    CheckPropViewNN,
                    CheckExPropView,
                    CheckExPropViewNL,
                    CheckExPropViewNN,
                    )

import logging
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


class Preferences(Qt.QObject):
    """ preferences tab """

    #: (:class:`taurus.qt.Qt.pyqtSignal`) server changed signal
    serverChanged = Qt.pyqtSignal()
    #: (:class:`taurus.qt.Qt.pyqtSignal`) layout changed signal
    layoutChanged = Qt.pyqtSignal(str, str)
    #: (:class:`taurus.qt.Qt.pyqtSignal`) dirty signal
    dirty = Qt.pyqtSignal()

    def __init__(self, ui, state=None):
        """ constructor

        :param ui: ui instance
        :type ui: :class:`taurus.qt.qtgui.util.ui.__UI`
        :param state: server state
        :type state: :class:`nxsselector.ServerState.ServerState`
        """
        Qt.QObject.__init__(self)
        #: (:class:`taurus.qt.qtgui.util.ui.__UI`) ui instance
        self.ui = ui
        #: (:class:`nxsselector.ServerState.ServerState`) server state
        self.state = state

        #: (:obj:`list` <:obj:`str`>) \
        #:     list of nested frames/columns/groups configuration frame layout
        self.frameshelp = [
            str(
                '[[[["Components",1],["Timers",5]]],[[["Counters", 4]],'
                + '[["ADC",3]],[["VFC",7]]],'
                + '[[["MCA/SCA",6],["Additional",0]]]]'),
            str('[[[["Components",1],["Timers",5]]],[[["Counters", 4]],'
                + '[["ADC",3]]],[[["MCA/SCA",6],["Devices",0]]]]'),
            str('[[[["Counters", 4]],[["Channels",0]]],'
                + '[[["MCAs", 2],["Misc",1]]],[[["ADC",3]]]]'),
            str('[[[["Devices", 0]]],[[["MCAs", 2],["Misc",1]]]]'),
            str(
                '[[[["Counters1", 0], ["Counters2", 2]], [["VCounters", 3]]],'
                + '[[["MCAs", 1], ["SCAs", 4]]], [[["Misc", 5] ]]]'),
            str('[[[["My Channels", 0]]],[[["My Components", 1]]]]'),
            str('')]
        #: (:obj:`list` <:class:`taurus.qt.str`>) \
        #:     list of configuration group content,
        #:      "<id>":[<pattern1>, <pattern2>, ...]
        self.mgroupshelp = [
            str('{ "3":["*_adc*"], "4":["*_c*"],'
                + '"5":["*_t*"], "6":["*_mca*"],'
                + '"7":["*_vfc*"]}'),
            str('{ "3":["exp_adc*"], "4":["exp_c*"],'
                + '"5":["exp_t*"], "6":["exp_mca*"],'
                + '["sca_exp_*"]}'),
            str('{"2":["mca8701*"]] , "3":["exp_adc*"],'
                + ' "4":["exp_c*"]}'),
            str('{"2":["ct01", "ct02"], "5":["appscan"]}'),
            str('')]
        #: (:obj:`list` <:obj:`str`>) \
        #:     list of configuration server
        self.serverhelp = [
            str(self.state.server)]

        #: (:obj:`str`) defualt group configuration
        self.__mgroups = str(self.mgroupshelp[0])
        #: (:obj:`str`) defualt frame configuration
        self.__frames = str(self.frameshelp[0])

        #: (:obj:`dict` <:obj:`str`, :class:`taurus.qt.Qt.QWidget`>) \
        #:    available element views
        self.views = {
            "CentralCheckBoxes (A)": CheckerView,
            "CheckBoxes (A)": LeftCheckerView,
            "Tables": TableView,
            "Columns": OneTableView,
            "CentralRadioButtons (A)": RadioView,
            "RadioButtons (A)": LeftRadioView,
            "Buttons (A)": ButtonView,
            "CentralCheckBoxes (N)": CheckerViewNL,
            "CheckBoxes (N)": LeftCheckerViewNL,
            "CentralRadioButtons (N)": RadioViewNL,
            "RadioButtons (N)": LeftRadioViewNL,
            "Buttons (N)": ButtonViewNL,
            "CentralCheckBoxes": CheckerViewNN,
            "CheckBoxes": LeftCheckerViewNN,
            "CentralRadioButtons": RadioViewNN,
            "RadioButtons": LeftRadioViewNN,
            "Buttons": ButtonViewNN,
            "CheckBoxes Dis (A)": CheckDisView,
            "CheckBoxes Prop (A)": CheckPropView,
            "CheckBoxes Synch (A)": CheckExPropView,
            "RadioButtons Dis (A)": RadioDisView,
            "CheckBoxes Dis (N)": CheckDisViewNL,
            "CheckBoxes Prop (N)": CheckPropViewNL,
            "CheckBoxes Synch (N)": CheckExPropViewNL,
            "RadioButtons Dis (N)": RadioDisViewNL,
            "CheckBoxes Dis": CheckDisViewNN,
            "CheckBoxes Prop": CheckPropViewNN,
            "CheckBoxes Synch": CheckExPropViewNN,
            "RadioButtons Dis": RadioDisViewNN,
            "CheckBoxes Dis (U)": CheckerLabelViewNN,
        }

        #: (:obj:`int`) maximum size of the helper for lineedits
        self.maxHelp = 10
        #: (:obj:`str`) layout file path
        self.layoutFile = os.getcwd()
        #: (:obj:`list` <:obj:`taurus.qt.Qt.QCompleter`>) \
        #:    qcompleter list
        self.completers = []

        #: (:obj:`bool`) if siganal connected
        self.connected = False
        self.connectSignals()

    def __setmgroups(self, groups):
        """ setter for mgroups

        :param groups: group configuration
        :type groups: :obj:`str`
        """
        try:
            lgroups = json.loads(groups)
            for gr in lgroups.values():
                for i in range(len(gr)):
                    if isinstance(gr[i], list) and gr[i]:
                        gr[i] = gr[i][0]
            self.__mgroups = json.dumps(lgroups)
        except Exception:
            self.__mgroups = "{}"

    def __getmgroups(self):
        """ getter for mgroups

        :returns: group configuration
        :rtype: :obj:`str`
        """
        return self.__mgroups

    #: (:obj:`str`) the json data string
    mgroups = property(__getmgroups, __setmgroups,
                       doc='device groups')

    def __setframes(self, frames):
        """ setter for frame configuration

        :param groups: frame configuration
        :type groups: :obj:`str`
        """
        try:
            nml = []
            lframes = json.loads(frames)
            for fr in lframes:
                for cl in fr:
                    for rw in cl:
                        nm = int(rw[1])
                        if nm not in nml:
                            nml.append(nm)
                        else:
                            raise Exception("Duplicated id")
            self.__frames = frames
        except Exception as e:
            print(str(e))

    def __getframes(self):
        """ getter for frame configuration

        :returns: group configuration
        :rtype: :obj:`str`
        """
        return self.__frames

    #: (:obj:`str`) the json data string
    frames = property(__getframes, __setframes,
                      doc='detector frames')

    def disconnectSignals(self):
        """ disconnects all signals
        """
        if self.connected:
            self.ui.devSettingsLineEdit.editingFinished.disconnect(
                self.on_devSettingsLineEdit_editingFinished)
            self.ui.devSettingsLineEdit.textEdited.disconnect(self.__dirty)
            self.ui.groupLineEdit.editingFinished.disconnect(
                self.on_layoutLineEdits_editingFinished)
            self.ui.frameLineEdit.editingFinished.disconnect(
                self.on_layoutLineEdits_editingFinished)

            self.ui.layoutButtonBox.button(
                Qt.QDialogButtonBox.Open).pressed.disconnect(self.layoutLoad)
            self.ui.layoutButtonBox.button(
                Qt.QDialogButtonBox.Save).pressed.disconnect(self.layoutSave)
        self.connected = False

    def connectSignals(self):
        """ connects all signals
        """
        if not self.connected:
            self.ui.devSettingsLineEdit.editingFinished.connect(
                self.on_devSettingsLineEdit_editingFinished)
            self.ui.devSettingsLineEdit.textEdited.connect(self.__dirty)
            self.ui.groupLineEdit.editingFinished.connect(
                self.on_layoutLineEdits_editingFinished)
            self.ui.frameLineEdit.editingFinished.connect(
                self.on_layoutLineEdits_editingFinished)

            self.ui.layoutButtonBox.button(
                Qt.QDialogButtonBox.Open).pressed.connect(self.layoutLoad)
            self.ui.layoutButtonBox.button(
                Qt.QDialogButtonBox.Save).pressed.connect(self.layoutSave)
        self.connected = True

    def __clearCompleters(self):
        """ clears completer list
        """
        self.ui.groupLineEdit.setCompleter(None)
        self.ui.devSettingsLineEdit.setCompleter(None)
        self.ui.frameLineEdit.setCompleter(None)
        DynamicTools.cleanupObjects(self.completers)

    def reset(self):
        """ resets whole Preferences
        """
        logger.debug("reset preferences")
        self.disconnectSignals()
        self.__clearCompleters()
        if self.ui.viewComboBox.count() != len(self.views.keys()):
            self.ui.viewComboBox.clear()
            self.ui.viewComboBox.addItems(sorted(self.views.keys()))
        completer = Qt.QCompleter(self.mgroupshelp, self.ui.preferences)
        self.ui.groupLineEdit.setCompleter(completer)
        self.completers.append(completer)
        completer = Qt.QCompleter(self.serverhelp, self.ui.preferences)
        self.ui.devSettingsLineEdit.setCompleter(completer)
        self.completers.append(completer)
        completer = Qt.QCompleter(self.frameshelp, self.ui.preferences)
        self.ui.frameLineEdit.setCompleter(completer)
        self.completers.append(completer)
        self.updateForm()
        self.connectSignals()
        logger.debug("reset preferences ended")

    def changeServer(self, ask=True):
        """ changes the selector server

        :param ask: ask  for confirmation from the user
        :type ask: :obj:`bool`
        """
        self.disconnectSignals()
        logger.debug("server changing")
        server = str(self.ui.devSettingsLineEdit.text())
        logger.debug("from %s to  %s" % (self.state.server, server))
        replay = Qt.QMessageBox.Yes
        if not server or server != self.state.server:
            if ask:
                replay = Qt.QMessageBox.question(
                    self.ui.preferences,
                    "Setting server has changed.",
                    "Changing server will cause loosing the current data. "
                    " Are you sure?",
                    Qt.QMessageBox.Yes | Qt.QMessageBox.No)
            if replay == Qt.QMessageBox.Yes:
                try:
                    if server == 'module':
                        self.state.server = ''
                        self.state.setServer()
                        self.state.fetchSettings()
                        self.addHint(server, self.serverhelp)
                    else:
                        if not server:
                            self.state.findServer()
                            self.state.setServer()
                            self.state.fetchSettings()
                        dp = PyTango.DeviceProxy(server)
                        if dp.info().dev_class == 'NXSRecSelector':
                            self.state.server = str(server)
                            self.state.setServer()
                            self.state.fetchSettings()
                            self.addHint(server, self.serverhelp)
                except Exception:
                    self.reset()
                self.connectSignals()
                self.serverChanged.emit()
            else:
                self.ui.devSettingsLineEdit.setText(str(
                    self.state.server if self.state.server else 'module'))
        self.connectSignals()

    @Qt.pyqtSlot()
    def on_devSettingsLineEdit_editingFinished(self):
        """ changes  selector server
        """
        logger.debug("on_devSettingsLineEdit_editingFinished")
        self.changeServer()
        logger.debug("server changed")

    def addHint(self, string, hints):
        """ adds hint

        :param string: string hint
        :type string: :obj:`str`
        :param hints: a list of hints
        :type hints: :obj:`list` <:obj:`str`>
        """
        qstring = str(string)
        if qstring not in hints:
            hints.insert(0, string)
        if self.maxHelp < len(hints):
            hints.pop(len(hints) - 1)

    @Qt.pyqtSlot()
    def on_layoutLineEdits_editingFinished(self):
        """ updates layout according to new frame and group configurations
        """
        logger.debug("on_layoutLineEdit_editingFinished")
        self.disconnectSignals()

        groups = str(self.ui.groupLineEdit.text())
        frames = str(self.ui.frameLineEdit.text())
        try:
            if not frames:
                frames = '[]'
            oldframes = self.frames
            self.frames = frames
            self.addHint(self.frames, self.frameshelp)

            if not groups:
                groups = '{}'
            mgroups = json.loads(groups)

            if isinstance(mgroups, dict):
                oldgroups = self.mgroups
                self.mgroups = groups
                self.addHint(groups, self.mgroupshelp)

                if oldframes != self.frames or oldgroups != self.mgroups:
                    self.connectSignals()
                    self.layoutChanged.emit(
                        str(frames), str(groups))
        except Exception as e:
            text = MessageBox.getText("Problem in setting layout")
            MessageBox.warning(
                self.ui.preferences,
                "NXSelector: Error during settings layout",
                text, str(e))
            logger.debug(str(e))
            self.reset()
        self.connectSignals()

    def updateForm(self):
        """ updates the form according to the class variables
        """
        self.ui.devSettingsLineEdit.setText(
            self.state.server if self.state.server else 'module')
        self.ui.groupLineEdit.setText(self.mgroups)
        self.ui.frameLineEdit.setText(self.frames)

    def apply(self):
        """ does nothing"""

    @Qt.pyqtSlot()
    def layoutLoad(self):
        """ loads a layout from the file
        """
        filename = str(
            Qt.QFileDialog.getOpenFileName(
                self.ui.preferences,
                "Load Layout",
                self.layoutFile,
                "JSON files (*.json);;All files (*)"))
        logger.debug("loading profile from %s" % filename)
        if filename:
            self.layoutFile = filename
            jprof = open(filename).read()
            try:
                profile = json.loads(jprof)
                if isinstance(profile, dict):
                    if "server" in profile.keys():
                        self.ui.devSettingsLineEdit.setText(
                            str(profile["server"]))
                        self.on_devSettingsLineEdit_editingFinished()
                    if "frames" in profile.keys():
                        self.ui.frameLineEdit.setText(
                            str(profile["frames"]))
                        self.on_layoutLineEdits_editingFinished()
                    if "groups" in profile.keys():
                        self.ui.groupLineEdit.setText(
                            str(profile["groups"]))
                        self.on_layoutLineEdits_editingFinished()
                    if "rowMax" in profile.keys():
                        self.ui.rowMaxSpinBox.setValue(
                            int(profile["rowMax"]))

            except Exception as e:
                text = MessageBox.getText("Problem in reading the layout")
                MessageBox.warning(
                    self.ui.preferences,
                    "NXSelector: Error during reading the file",
                    text, str(e))

    @Qt.pyqtSlot()
    def __dirty(self):
        """ sets dirty flag """
        self.dirty.emit()

    @Qt.pyqtSlot()
    def layoutSave(self):
        """ loads the current layout into a file
        """
        try:
            filename = str(Qt.QFileDialog.getSaveFileName(
                self.ui.storage,
                "Save Layout",
                self.layoutFile,
                "JSON files (*.json);;All files (*)"))
            logger.debug("saving profile to %s" % filename)
            if filename:
                if (len(filename) < 4 or filename[-4] != '.') and \
                        not (len(filename) > 5 and filename[-5] == '.'):
                    filename = filename + '.json'

                self.layoutFile = filename
                profile = {}
                profile["server"] = str(self.ui.devSettingsLineEdit.text())
                profile["frames"] = str(self.ui.frameLineEdit.text())
                profile["groups"] = str(self.ui.groupLineEdit.text())
                profile["rowMax"] = self.ui.rowMaxSpinBox.value()
                jprof = json.dumps(profile)
                with open(filename, 'w') as myfile:
                    myfile.write(jprof)
        except Exception as e:
            text = MessageBox.getText("Problem in saving the layout")
            MessageBox.warning(
                self.ui.preferences,
                "NXSelector: Error during saving the file",
                text, str(e))



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# state of sardana recorder server

""" state of recorder server """

import PyTango
import json
import time
import subprocess

try:
    from taurus.external.qt import Qt
except Exception:
    from taurus.qt import Qt


import logging
#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


class Checker(object):
    """ compare configuation methods
    """

    def __init__(self):
        """ constructor
        """

        #: (:obj:`list`<:obj:`str`>) special keys for sorted json lists
        self.jsortedlists = ['PreselectingDataSources']
        #: (:obj:`list`<:obj:`str`>) special keys for json dicts
        self.jdicts = ['DataSourceSelection']

    def compDict(self, dct, dct2, sort=False):
        """ compare two configuration dictionaries

        :param dct: first configuration dictionary
        :type dct: `dict`<:obj:`any`, :obj:`any`>
        :param dct2: second configuration dictionary
        :type dct2: `dict`<:obj:`any`, :obj:`any`>
        :param sort: if use special keys
        :type sort: :obj:`bool`
        :returns: if configuration are equal
        :rtype: :obj:`bool`
        """

        if not isinstance(dct, dict):
            return False
        if not isinstance(dct2, dict):
            return False
        if len(dct.keys()) != len(dct2.keys()):
            return False
        for k, v in dct.items():
            if k not in dct2.keys():
                return False
            if isinstance(v, dict):
                return self.compDict(v, dct2[k])
            elif sort and k in self.jsortedlists:
                l1 = sorted(json.loads(v))
                l2 = sorted(json.loads(dct2[k]))
                if l1 != l2:
                    return False
            elif sort and k in self.jdicts:
                d1 = json.loads(v)
                d2 = json.loads(dct2[k])
                return self.compDict(d1, d2)
            else:
                if v != dct2[k]:
                    return False
        return True


class SynchThread(Qt.QThread):
    """ thread with server command
    """

    #: (:class:`taurus.qt.Qt.pyqtSignal`) dirty signal
    scanidchanged = Qt.pyqtSignal()

    #: (:class:`taurus.qt.Qt.pyqtSignal') mg configuration changed
    mgconfchanged = Qt.pyqtSignal()

    def __init__(self, serverstate, server, mutex):
        """constructor
        :param serverstate: ServerState
        :type serverstate: :class:`ServerState`
        :param server: server name
        :type server: :obj:`str`
        """
        Qt.QThread.__init__(self, serverstate)
        #: (:obj:`bool`) server is running
        self.running = True

        #: (:class:`taurus.qt.Qt.QMutex`) thread mutex
        self.mutex = mutex

        #: (:obj:`str`) server name
        self.server = str(serverstate.server) if serverstate.server else None

        #: (:obj:`str`) server name
        self.__serverstate = serverstate

        #: (:obj:`int`) last scan id
        self.__lastscanid = 0
        #: (:obj:`str`) last mntgrp configuration
        self.__lastmg = ""
        #: (:obj:`str`) last profile configuration
        self.__lastprof = ""

        self.__dp = None
        if self.server and self.server != 'module':
            self.__dp = PyTango.DeviceProxy(self.server)
            self.__dp.set_source(PyTango.DevSource.DEV)
            self.__lastscanid = self.__dp.scanID
            self.__lastmg = self.__dp.mntGrpConfiguration()
            self.__lastprof = self.__dp.profileConfiguration

    def restart(self):
        with Qt.QMutexLocker(self.mutex):
            self.server = str(self.__serverstate.server) \
                if self.__serverstate.server else None
        if self.server and self.server != 'module':
            self.__dp = PyTango.DeviceProxy(self.server)
            self.__dp.set_source(PyTango.DevSource.DEV)
            self.__lastscanid = self.__dp.scanID
            self.__lastmg = self.__dp.mntGrpConfiguration()
            self.__lastprof = self.__dp.profileConfiguration
        self.running = True
        self.start()

    def run(self):
        """ runs synch thread
        """
        insynch = True
        checker = Checker()
        while insynch:
            self.sleep(5)
            try:
                if not Qt or not self.__dp:
                    break
                scanid = self.__dp.scanID
                mg = self.__dp.mntGrpConfiguration()
                prof = self.__dp.profileConfiguration
                with Qt.QMutexLocker(self.mutex):
                    if not self.running:
                        insynch = False
                if self.__lastscanid != scanid and insynch:
                    self.scanidchanged.emit()
                    self.__lastscanid = scanid
                if (self.__lastmg != mg or self.__lastprof != prof) \
                   and insynch:
                    m0 = json.loads(self.__lastmg)
                    m1 = json.loads(mg)
                    if not checker.compDict(m0, m1):
                        self.mgconfchanged.emit()
                    else:
                        p0 = json.loads(self.__lastprof)
                        p1 = json.loads(prof)
                        if not checker.compDict(p0, p1, True):
                            self.mgconfchanged.emit()

                    self.__lastmg = mg
                    self.__lastprof = prof
            except Exception:
                """ what is wrong """


class ServerState(Qt.QObject):
    """ state of recorder server """

    #: (:class:`taurus.qt.Qt.pyqtSignal`) server changed signal
    serverChanged = Qt.pyqtSignal()

    def __init__(self, server=None):
        """ constructor

        :param server: selector server name
        :type server: :obj:`str`
        """
        Qt.QObject.__init__(self)

        #: (:obj:`str`) selector server name
        self.server = None

        #: (:class:`taurus.qt.Qt.QMutex`) thread mutex
        self.mutex = Qt.QMutex()

        #: (:class:`PyTango.Database`) tango database instance
        self.__db = PyTango.Database()
        #: (:class:`PyTango.DeviceProxy`) selector server device proxy
        self.__dp = None

        #: (:obj:`int`) timeout
        self.__timeout = 25000

        #: (:obj:`dict` < :obj:`str`, `any`>) \
        #:     profile server configuration
        self.__conf = {}

        self.findServer(server)

        #: (:obj:`list`<:obj:`str`>) tango database
        self.errors = []

        #: (:obj:`str`) scan directory
        self.scanDir = None
        #: (:obj:`list`<:obj:`str`>) scan file names
        self.scanFile = []
        #: (:obj:`int`) scan id
        self.scanID = 0

        #: (:obj:`list`<:obj:`str`>) timers
        self.timers = None
        #: (:obj:`str`) measurement group name
        self.mntgrp = None
        #: (:obj:`str`) door  device name
        self.door = None

        #: (:obj:`str`) configuration device name
        self.configDevice = None
        #: (:obj:`str`) writer device name
        self.writerDevice = None

        #: (:obj:`bool`) append entry in one file
        self.appendEntry = None

        #: (:obj:`bool`) use dynamic components
        self.dynamicComponents = True
        #: (:obj:`bool`) use dynamic links
        self.dynamicLinks = None
        #: (:obj:`str`) default dynamic nexus path
        self.dynamicPath = None

        #: (:obj:`dict` <:obj:`str` , :obj:`bool` or `None`>) \
        #:    datasource selection
        self.dsgroup = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`str`>) device labels
        self.labels = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`dict` <:obj:`str` , `any`>>) \
        #:     device properties
        self.properties = {}
        #: (:obj:`list`<:obj:`str`>) list of non-plotted devices
        self.nodisplay = []
        #: (:obj:`dict` <:obj:`str` , :obj:`bool` or `None`>) \
        #:    detector component selection
        self.cpgroup = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`bool` or `None`>) \
        #:    description component selection
        self.acpgroup = {}
        #: (:obj:`list`<:obj:`str`>) selected description components
        self.acplist = []
        #: (:obj:`list`<:obj:`str`>) available timers
        self.atlist = []
        #: (:obj:`list`<:obj:`str`>) mandatory components
        self.mcplist = []
        #: ( [:obj:`dict` <:obj:`str`, :obj:`dict` <:obj:`str`, \
        #:        :obj:`list` <(:obj:`str`, :obj:`str`, :obj:`str`, \
        #:        :obj:`str`, :obj:`list` <:obj:`int`>)> > > ] ) \
        #: element description
        self.description = []
        #: (:obj:`list`<:obj:`str`>) available components
        self.avcplist = []
        #: (:obj:`list`<:obj:`str`>) available datasources
        self.avdslist = []
        #: (:obj:`list`<:obj:`str`>) available measurement groups
        self.avmglist = []
        #: (:obj:`dict` <:obj:`str`, :obj:`list` <:obj:`str` >>) \
        #:     variable components
        self.vrcpdict = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`list` <:obj:`str` >>) \
        #:     component variables
        self.cpvrdict = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`bool` or `None`>) \
        #:    init (descriptive) datasource selection
        self.idsgroup = {}
        #: (:obj:`list`<:obj:`str`>) administrator data key names
        self.admindata = []

        #: (:obj:`list`<:obj:`str`>) ordered pool channels
        self.orderedchannels = []

        #: (:obj:`dict` <:obj:`str` , :obj:`str`>) configuration variables
        self.configvars = {}
        #: (:obj:`dict` <:obj:`str` , `any`>) (name, value) \
        #:     user data dictionary
        self.datarecord = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`str`>) full device names
        self.fullnames = {}

        #: (:obj:`dict` <:obj:`str` , :obj:`bool`>) label links
        self.labellinks = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`bool`>) label canfail flags
        self.labelcanfailflags = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`str`>) label nexus paths
        self.labelpaths = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`list`< :obj:`int`> >) \
        #:     label data shapes
        self.labelshapes = {}
        #: (:obj:`dict` <:obj:`str` , :obj:`str`>) label nexus types
        self.labeltypes = {}

        #: (:obj:`list`<:obj:`str`>) error list
        self.errors = []
        #: (:obj:`bool`) no timer restriction flag
        self.notimerresctriction = False

        self.ddsdirty = True
        self.__ddsbackup = {}

        try:
            self.setServer()
            if self.server:
                self.__dp.ping()
        except Exception:
            self.server = None
            raise
        logger.debug("DP %s" % type(self.__dp))

        self.recorder_names = ['serialno', 'end_time', 'start_time',
                               'point_nb', 'timestamps', 'scan_title',
                               'filename'
                               ]
        self.channelprops = ["nexus_path", "link", "shape", "label",
                             "data_type", "canfail"]
        self.extrachannelprops = ["synchronizer", "synchronization"]
        self.synchthread = SynchThread(self, self.server, self.mutex)

    def __grepServer(self):
        """ provides the local selector server device name

        :returns: selector server device name
        :rtype: :obj:`str`
        """
        server = None
        try:
            pipe = subprocess.Popen("ps -ef | grep 'NXSRecSelecto'",
                                    stdout=subprocess.PIPE,
                                    shell=True, bufsize=10000).stdout
            res = pipe.read().split("\n")
            cres = [r for r in res if 'NXSRecSelector' in r]
            mi = 0
            if len(cres) > 0:
                command = cres[0].split()
                for i in range(len(command)):
                    if 'NXSRecSelector' in command[i]:
                        mi = i
                        break
                if len(command) > mi + 1:
                    instance = command[mi + 1]
                    server = self.__db.get_device_class_list(
                        "NXSRecSelector/%s" % instance).value_string[2]
        except Exception:
            pass
        return server

    def findServer(self, server=None):
        """  sets the existing NXSRecSelector server

        :param server: selector server device name
        :type server: :obj:`str`
        """
        if server is None:
            servers = self.__db.get_device_exported_for_class(
                "NXSRecSelector").value_string
            if len(servers):
                if len(servers) > 1:
                    gserver = self.__grepServer()
                    if gserver in servers:
                        self.server = str(gserver)
                else:
                    self.server = str(servers[0])
            else:
                self.server = None
        elif not server:
            self.server = None
        else:
            self.server = str(server)
        self.updateServerShared()

    def updateServerShared(self):
        with Qt.QMutexLocker(self.mutex):
            if hasattr(self, "synchthread"):
                if self.server != self.synchthread.server:
                    self.serverChanged.emit()

    def __fetchConfiguration(self):
        """ fetches from the server the current profile configuration
        """
        if not self.__dp:
            self.setServer()
        if not self.server:
            self.__dp.exportEnvProfile()
        self.__conf = json.loads(self.__dp.profileConfiguration)

    def createDataSources(self, datasources):
        """ creates new datasources
        :param datasources: datasources to add { name: source }
        :type datasources: :obj:`dict` <:obj:`str`, :obj:`str``>
        """
        if not self.__dp:
            self.setServer()
        self.__command(self.__dp, "createDataSources",
                       str(json.dumps(datasources)))
        self.__fetchConfiguration()
        self.avdslist = self.__getList("availableDataSources")
        self.dsdescription = self.__getList(
            "dataSourceDescription", argin=self.avdslist)

    def fetchErrors(self):
        """ fetches from the server the description errors
        """
        if not self.__dp:
            self.setServer()

        dc = "Error: Cannot read descriptionErrors"
        if self.server:
            error = True
            maxcount = 10
            while error and maxcount:
                try:
                    dc = self.__dp.read_attribute("descriptionErrors").value
                    error = False
                except Exception as e:
                    logger.warning(str(e))
                maxcount -= 1
        else:
            dc = getattr(self.__dp, "descriptionErrors")
        logger.debug(dc)
        self.errors = []
        if dc:
            if isinstance(dc, (list, tuple)):
                self.errors = dc
        logger.debug(" %s = %s" % ("descriptionErrors", self.errors))
        return self.errors

    def setProperties(self):
        """ sets label properties from properties
        """
        if "label" in self.properties:
            self.labels = self.properties["label"]
        else:
            self.labels = {}
        if "link" in self.properties:
            self.labellinks = self.properties["link"]
        else:
            self.labellinks = {}
        if "canfail" in self.properties:
            self.labelcanfailflags = self.properties["canfail"]
        else:
            self.labelcanfailflags = {}
        if "nexus_path" in self.properties:
            self.labelpaths = self.properties["nexus_path"]
        else:
            self.labelpaths = {}
        if "shape" in self.properties:
            self.labelshapes = self.properties["shape"]
        else:
            self.labelshapes = {}
        if "data_type" in self.properties:
            self.labeltypes = self.properties["data_type"]
        else:
            self.labeltypes = {}
        if "__controllers__" in self.properties:
            self.controllers = self.properties["__controllers__"]
        else:
            self.controllers = {}
        if "__triggergatelist__" in self.properties:
            self.triggergatelist = \
                self.properties["__triggergatelist__"]
        else:
            self.triggergatelist = []

    def getProperties(self):
        """ sets properties from label properties
        """
        self.properties["label"] = self.labels
        self.properties["link"] = self.labellinks
        self.properties["canfail"] = self.labelcanfailflags
        self.properties["nexus_path"] = self.labelpaths
        self.properties["shape"] = self.labelshapes
        self.properties["data_type"] = self.labeltypes

    def fetchSettings(self):
        """ fetches configuration setting from server
        """
        self.__fetchConfiguration()

        self.cpgroup = self.__importDict("ComponentSelection")
        self.dsgroup = self.__importDict("DataSourceSelection")
        self.acpgroup = self.__importDict("ComponentPreselection")
        self.properties = self.__importDict("ChannelProperties")
        self.setProperties()
        self.datarecord = self.__importDict("UserData")
        self.configvars = self.__importDict("ConfigVariables")

        self.nodisplay = self.__importList("UnplottedComponents", True)
        self.orderedchannels = self.__importList("OrderedChannels", True)
        self.idsgroup = self.__importDict("DataSourcePreselection")

        self.avcplist = self.__getList("availableComponents")
        self.avdslist = self.__getList("availableDataSources")
        self.dsdescription = self.__getList(
            "dataSourceDescription", argin=self.avdslist)
        self.avmglist = self.__getList("availableMntGrps")
        self.mcplist = self.__getList("mandatoryComponents")

        self.acplist = self.__getList("preselectedComponents")
        self.atlist = self.__getList("availableTimers")
        self.description = self.__getList("componentDescription", True)
        self.ddsdirty = True
        self.mutedChannels = self.__getList("mutedChannels")

        self.vrcpdict = self.__getDict("variableComponents")
        self.fullnames = self.__getDict("fullDeviceNames")
        self.admindata = self.__getList("administratorDataNames")

        self.motors = self.__getList(
            "poolElementNames", argin='MotorList')
        self.acqchannels = self.__getList(
            "poolElementNames", argin='AcqChannelList')
        self.ioregisters = self.__getList(
            "poolElementNames", argin='IORegisterList')

        self.__fetchFileData()
        self.fetchEnvData()
        if self.notimerresctriction:
            # old version to check
            self.atlist = list(set(self.atlist) | set(self.timers))
        else:
            if self.timers:
                self.atlist = list(set(self.atlist) | set([self.timers[0]]))
            self.timers = [tm for tm in self.timers if tm in self.atlist]

        self.cpvrdict = {}
        for vr, cps in self.vrcpdict.items():
            for cp in cps:
                if cp not in self.cpvrdict.keys():
                    self.cpvrdict[cp] = set()
                self.cpvrdict[cp].add(vr)

    def __fetchFileData(self):
        """ fetches file data configuration from the server
        """
        self.timers = self.__importList("Timer", True)
        self.mntgrp = str(self.__importData("MntGrp"))
        try:
            self.door = str(self.__loadData("door"))
        except Exception:
            self.storeData("door", "")
            self.door = str(self.__loadData("door"))

        self.configDevice = str(self.__loadData("configDevice"))
        self.writerDevice = str(self.__importData("WriterDevice"))

        self.appendEntry = self.__importData("AppendEntry")
        self.dynamicLinks = self.__importData("DefaultDynamicLinks")
        self.dynamicPath = str(self.__importData("DefaultDynamicPath"))

    def fetchEnvData(self, params=None):
        """ fetches scan variables from the server
        """
        params = params or {"ScanDir": "scanDir",
                            "ScanFile": "scanFile",
                            "ScanID": "scanID"}

        if not self.__dp:
            self.setServer()

        jvalue = self.__command(self.__dp, "scanEnvVariables")
        value = json.loads(jvalue)

        for var, attr in params.items():
            if var in value.keys():
                setattr(self, attr, value[var])
        logger.debug("fetch Env: %s" % (jvalue))

    def __storeEnvData(self):
        """ stores scan variables on the server
        """
        params = {"ScanDir": "scanDir",
                  "ScanFile": "scanFile",
                  "NeXusSelectorDevice": "server",
                  # "ScanID": "scanID"
                  }

        if not self.__dp:
            self.setServer()

        value = {}
        for var, attr in params.items():
            value[var] = getattr(self, attr)
        jvalue = json.dumps(value)
        self.scanID = self.__command(self.__dp, "setScanEnvVariables", jvalue)
        logger.debug("Store Env: %s" % (jvalue))

    def __storeFileData(self):
        """ stores file data configuration on the server
        """

        self.storeData("configDevice", self.configDevice)
        self.storeData("door", self.door)

        self.__exportData("ConfigDevice", self.configDevice)
        self.__exportData("Door", self.door)
        self.__exportData("MntGrp", self.mntgrp)

        self.__exportData("WriterDevice", self.writerDevice)
        self.__exportList("Timer", self.timers)
        self.__exportData("AppendEntry", self.appendEntry)
        self.__exportData("DynamicComponents", self.dynamicComponents)
        self.__exportData("DefaultDynamicLinks", self.dynamicLinks)
        self.__exportData("DefaultDynamicPath", self.dynamicPath)

    def storeGroups(self):
        """ stores selection group settings on the server"""
        if not self.__dp:
            self.setServer()
        self.__exportDict("DataSourceSelection", self.dsgroup)
        self.__exportDict("ComponentSelection", self.cpgroup)
        self.__exportDict("ComponentPreselection", self.acpgroup)
        self.__exportDict("DataSourcePreselection", self.idsgroup)
        self.__storeConfiguration()

    def storeSettings(self):
        """ stores all settings on the server"""
        if not self.__dp:
            self.setServer()
        self.__storeEnvData()
        self.__storeFileData()
        self.__exportDict("DataSourceSelection", self.dsgroup)
        self.__exportDict("ComponentSelection", self.cpgroup)
        self.__exportDict("ComponentPreselection", self.acpgroup)
        self.__exportDict("DataSourcePreselection", self.idsgroup)
        self.getProperties()
        self.__exportDict("ChannelProperties", self.properties)
        self.__exportList("UnplottedComponents", self.nodisplay)
        self.__exportList("OrderedChannels", self.orderedchannels)
        self.__exportDict("UserData", self.datarecord)
        self.__exportDict("ConfigVariables", self.configvars)
        if not self.server:
            self.__dp.exportEnvProfile()
        self.__storeConfiguration()

    def __storeConfiguration(self):
        """ stores profile configuration on the server"""
        if not self.__dp:
            self.setServer()
        self.__dp.profileConfiguration = str(json.dumps(self.__conf))
        if not self.server:
            self.__dp.exportEnvProfile()

    def fetchMntGrp(self):
        """ fetches mntgrp and profile from the server
        """
        if not self.__dp:
            self.setServer()
        self.__command(self.__dp, "fetchProfile")

    def switchMntGrp(self):
        """ switches mntgrp and profile on the server
        """
        if not self.__dp:
            self.setServer()
        self.__command(self.__dp, "switchProfile")

    def updateMntGrp(self):
        """ updates mntgrp on the macroserver/pool
        """
        if not self.mntgrp:
            raise Exception("ActiveMntGrp not defined")
        if not self.scanFile:
            raise Exception("ScanFile not defined")
        if not self.scanDir:
            raise Exception("ScanDir not defined")
        self.storeSettings()
        mgconf = self.__command(self.__dp, "updateMntGrp")
        conf = {}
        conf['MntGrpConfigs'] = {}
        conf['ActiveMntGrp'] = self.mntgrp
        conf['MntGrpConfigs'][self.mntgrp] = json.loads(mgconf)
        return conf

    def isMntGrpChanged(self):
        """ checks if measurement group has changed

        :returns: if measurement group has changed
        :rtype: :obj:`bool`
        """
        if not self.__dp:
            self.setServer()
        mgconf = json.loads(self.__command(self.__dp, "mntGrpConfiguration"))
        locmgconf = self.__importDict("MntGrpConfiguration")
        checker = Checker()
        if checker.compDict(mgconf, locmgconf):
            pconf = json.loads(self.__dp.profileConfiguration)
            locpconf = self.__conf
            return checker.compDict(pconf, locpconf, True)
        return False

    def importMntGrp(self):
        """ imports mntgrp from sardana
        """
        if not self.__dp:
            self.setServer()
        return self.__command(self.__dp, "importMntGrp")

    def createConfiguration(self):
        """ creates the NeXus Writer configuration

        :returns: NeXus writer configuration
        :rtype: :obj:`str`
        """
        if not self.__dp:
            self.setServer()
        return self.__command(self.__dp, "createWriterConfiguration", [])

    def deleteMntGrp(self, name):
        """ deletes the profile and the corresponding measurement group

        :param name: measurement group name
        :type name: :obj:`str`
        """
        if not self.__dp:
            self.setServer()
        self.__command(self.__dp, "deleteProfile", str(name))
        self.avmglist = self.__getList("availableMntGrps")
        if self.avmglist:
            self.mntgrp = self.avmglist[0]
            self.storeData("mntGrp", self.mntgrp)
            self.fetchMntGrp()

    def mntGrpConfiguration(self):
        """ provides measurement group configuration with its name

        :returns: measurement group configuration with its name
        :rtype: :obj:`str`
        """
        mgconf = self.__command(self.__dp, "mntGrpConfiguration")
        conf = {}
        conf['MntGrpConfigs'] = {}
        conf['ActiveMntGrp'] = self.mntgrp
        conf['MntGrpConfigs'][self.mntgrp] = json.loads(mgconf)
        return json.dumps(conf)

    def lastMntGrpConfiguration(self):
        """ provides measurement group configuration with its name

        :returns: measurement group configuration with its name
        :rtype: :obj:`str`
        """
        conf = {}
        conf['MntGrpConfigs'] = {}
        conf['ActiveMntGrp'] = self.mntgrp
        conf['MntGrpConfigs'][self.mntgrp] = \
            self.__importDict("MntGrpConfiguration")
        return json.dumps(conf)

    def getConfiguration(self):
        """ provides profile configuration

        :returns: profile configuration
        :rtype: :obj:`str`
        """
        self.storeSettings()
        return self.__dp.profileConfiguration

    def setConfiguration(self, conf):
        """ sets profile configuration

        :param conf: profile configuration
        :type conf: :obj:`str`
        """
        self.__dp.profileConfiguration = conf
        self.__command(self.__dp, "updateMntGrp")
        self.fetchSettings()

    def resetDescriptions(self):
        """ resets description components to default values
        """
        if hasattr(self.__dp, "command_inout_asynch"):
            # aid = self.__dp.command_inout_asynch("PreselectComponents")
            # self.__wait(self.__dp)
            try:
                self.__command(self.__dp, "resetPreselectedComponents")
            except PyTango.CommunicationFailed as e:
                if e[-1].reason == "API_DeviceTimedOut":
                    self.__wait(self.__dp)
                else:
                    raise
        else:
            self.__command(self.__dp, "resetPreselectedComponents")

    def updateControllers(self):
        """ update description component selection accoriding
             to its device state
        """
        if hasattr(self.__dp, "command_inout_asynch"):
            # aid = self.__dp.command_inout_asynch("PreselectComponents")
            # self.__wait(self.__dp)
            try:
                self.__command(self.__dp, "PreselectComponents")
            except PyTango.CommunicationFailed as e:
                if e[-1].reason == "API_DeviceTimedOut":
                    self.__wait(self.__dp)
                else:
                    raise
        else:
            self.__command(self.__dp, "preselectComponents")

    def setServer(self):
        """ sets the selector server
        """
        if self.server:
            self.__dp = self.__openProxy(self.server)
            self.__dp.set_source(PyTango.DevSource.DEV)
            self.__dp.set_timeout_millis(self.__timeout)
            logger.debug("set server: %s:%s/%s" % (self.__dp.get_db_host(),
                                                   self.__dp.get_db_port(),
                                                   self.__dp.name()))
        else:
            from nxsrecconfig import Settings
            self.__dp = Settings.Settings()
        if not hasattr(self.__dp, "version") or \
           int(str(self.__dp.version).split(".")[0]) < 2:
            raise Exception("NXSRecSelector (%s) version below 2.0.0" %
                            (self.server or "module"))

    def isDoorFromMacroServer(self, door):
        """ checks if door is of the current MacroServer

        :param door: door name
        :type door: :obj:`str`
        :returns: if door is of the current MacroServer
        :rtype: :obj:`bool`
        """
        if not self.__dp:
            self.setServer()
        if hasattr(self.__dp, "macroServer"):
            ms = str(self.__dp.macroServer)
            if ms:
                if ':' not in ms and hasattr(self.__dp, "get_db_host"):
                    ms = "%s:%s/%s" % (self.__dp.get_db_host(),
                                       self.__dp.get_db_port(),
                                       ms)

                msp = self.__openProxy(ms)
                host = msp.get_db_host()
                port = msp.get_db_port()
                doors = msp.doorList
                if door in doors:
                    return True
                status = False
                for mdoor in doors:
                    if ':' in door and ':' not in mdoor:
                        if door == "%s:%s/%s" % (host, port, mdoor):
                            status = True
                            break
                    elif ':' not in door and ':' in mdoor:
                        if mdoor == "%s:%s/%s" % (host, port, door):
                            status = True
                            break
                return status

    @classmethod
    def __openProxy(cls, server):
        """ creates device proxy

        :param server: server name
        :type server: :obj:`str`
        :returns: server device proxy
        :rtype: :class:`PyTango.DeviceProxy`
        """
        proxy = PyTango.DeviceProxy(server)
        cls.__wait(proxy)
        return proxy

    @classmethod
    def __command(cls, server, command, *var):
        """ executes command on the server

        :param server: server instance
        :type server: :class:`PyTango.DeviceProxy` \
                      or 'nxsrecconfig.Settings.Settings'
        :param command: command name
        :type command: :obj:`str`
        :returns: command result
        :rtype: `any`

        """
        if not hasattr(server, "command_inout"):
            return getattr(server, command)(*var)
        elif var is None:
            return server.command_inout(command)
        else:
            return server.command_inout(command, *var)

    @classmethod
    def __wait(cls, proxy, counter=100):
        """ waits for server until server is not in running state

        :param proxy: server proxy
        :type proxy: :class:`PyTango.DeviceProxy`
        :param counter: maximum waiting timer in 0.01 sec
                        (without command execution)
        :type counter: :obj:`int`
        """
        found = False
        cnt = 0
        while not found and cnt < counter:
            if cnt > 1:
                time.sleep(0.01)
            try:
                if proxy.state() != PyTango.DevState.RUNNING:
                    found = True
            except PyTango.DevFailed:
                time.sleep(0.01)
                found = False
                if cnt == counter - 1:
                    raise
            cnt += 1

    def __importDict(self, name):
        """ imports a dictionary variable from the profile configuration

        :param name: record name
        :type name: :obj:`str`
        :returns: returns dictionary
        :rtype: :obj:`dict` <`any`, `any`>
        """
        dsg = self.__conf[name] if name in self.__conf else None
        res = {}
        if dsg:
            dc = json.loads(dsg)
            if isinstance(dc, dict):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __importList(self, name, encoded=False):
        """ imports a list variable from the profile configuration

        :param name: record name
        :type name: :obj:`str`
        :param encoded: if list should be encoded from JSON
        :type encoded: :obj:`bool`
        :returns: returns configuration list
        :rtype: :obj:`list` <`any`>
        """
        dc = self.__conf[name] if name in self.__conf else None
        logger.debug(dc)
        res = []
        if dc:
            if encoded:
                dc = json.loads(dc)
            if isinstance(dc, (list, tuple)):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __importData(self, name):
        """ imports a variable from the profile configuration

        :param name: record name
        :type name: :obj:`str`
        :returns: returns configuration variable
        :rtype: `any`
        """
        dc = self.__conf[name] if name in self.__conf else None
        logger.debug(dc)
        return dc

    def __loadDict(self, name):
        """ reads dictionary variable from the configuration server attribute

        :param name: attribute name
        :type name: :obj:`str`
        :returns: returns dictionary
        :rtype: :obj:`dict` <`any`, `any`>
        """

        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()
            dsg = self.__dp.read_attribute(name).value
        else:
            dsg = getattr(self.__dp, name)
        res = {}
        if dsg:
            dc = json.loads(dsg)
            if isinstance(dc, dict):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __exportDict(self, name, value):
        """ writes a dictionary variable into the profile configuration

        :param name: attribute name
        :type name: :obj:`str`
        :param value: returns dictionary
        :type value: :obj:`dict` <`any`, `any`>
        """
        self.__conf[name] = json.dumps(value)
        logger.debug(" %s = %s" % (name, value))

    def __exportList(self, name, value):
        """ writes a list variable into the profile configuration

        :param name: attribute name
        :type name: :obj:`str`
        :param value: returns dictionary
        :type value: :obj:`list` <`any`>
        """
        self.__conf[name] = json.dumps(value)
        logger.debug(" %s = %s" % (name, value))

    def storeData(self, name, value):
        """ stores data into the configuration server attribute

        :param name: attribute name
        :type name: :obj:`str`
        :param value: attribute value
        :type value: :obj:`list` <`any`>
        """
        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()

        if self.server:
            try:
                self.__dp.write_attribute(name, value)
            except PyTango.CommunicationFailed as e:
                if e[-1].reason == "API_DeviceTimedOut":
                    self.__wait(self.__dp)
                else:
                    raise
        else:
            setattr(self.__dp, name, value)
        logger.debug(" %s = %s" % (name, value))

    def __exportData(self, name, value):
        """ writes a variable into the profile configuration

        :param name: attribute name
        :type name: :obj:`str`
        :param value: returns dictionary
        :type value: `any`
        """
        self.__conf[name] = value
        logger.debug(" %s = %s" % (name, value))

    def __loadList(self, name, encoded=False):
        """ reads a list variable from the configuration server attribute

        :param name: attribute name
        :type name: :obj:`str`
        :param encoded: encoding should be used
        :type encoded: :obj:`bool`
        :returns: returns dictionary
        :rtype: :obj:`list` <`any`>
        """
        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()
            dc = self.__dp.read_attribute(name).value
        else:
            dc = getattr(self.__dp, name)
        logger.debug(dc)
        res = []
        if dc:
            if encoded:
                dc = json.loads(dc)
            if isinstance(dc, (list, tuple)):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __loadData(self, name):
        """ reads a attribute value from the configuration server

        :param name: attribute name
        :type name: :obj:`str`
        :returns: returns an attribute value
        :rtype: `any`
        """
        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()
            dc = self.__dp.read_attribute(name).value
        else:
            dc = getattr(self.__dp, name)

        logger.debug(dc)
        return dc

    def __getList(self, name, encoded=False, argin=None):
        """ returns a result list of the selection server command

        :param name: record name
        :type name: :obj:`str`
        :param encoded: if list should be encoded from JSON
        :type encoded: :obj:`bool`
        :param argin: input command argument
        :type argin: `any`
        :returns: returns the command result list
        :rtype: :obj:`list` <`any`>
        """
        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()
            if argin is None:
                dc = self.__dp.command_inout(name)
            else:
                dc = self.__dp.command_inout(name, argin)

        else:
            if argin is None:
                dc = getattr(self.__dp, name)()
            else:
                dc = getattr(self.__dp, name)(argin)

        logger.debug(dc)
        res = []
        if dc:
            if encoded:
                dc = json.loads(dc)
            if isinstance(dc, (list, tuple)):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __getDict(self, name):
        """ returns a result dictionary of the selection server command

        :param name: record name
        :type name: :obj:`str`
        :returns: returns the command result dictionary
        :rtype: :obj:`dict` <`any`, `any`>
        """
        if not self.__dp:
            self.setServer()
        if self.server:
            self.__dp.ping()
            dc = self.__dp.command_inout(name)
        else:
            dc = getattr(self.__dp, name)()

        logger.debug(dc)
        res = {}
        if dc:
            dc = json.loads(dc)
            if isinstance(dc, (dict)):
                res = dc
        logger.debug(" %s = %s" % (name, res))
        return res

    def __disableDataSources(self):
        """ provides disable datasources

        :returns: (disable datasources, ds component) dictionary
        :rtype: :obj:`dict` <:obj:`str`, :obj:`str`>
        """
        if not self.ddsdirty:
            return dict(self.__ddsbackup)

        res = self.description
        dds = {}
        for cpg in res:
            for cp, dss in cpg.items():
                if isinstance(dss, dict):
                    if cp in self.cplist \
                       or cp in self.mcplist \
                       or cp in self.dslist \
                       or cp in self.acplist:
                        for ds, values in dss.items():
                            for vl in values:
                                if len(vl) > 0 and vl[0] == 'STEP':
                                    if ds != cp:
                                        dds[ds] = cp
                                        break
        if self.timers:
            for timer in self.timers:
                if timer not in dds.keys():
                    dds[timer] = ''
        self.__ddsbackup = dict(dds)
        self.ddsdirty = False
        return dds

    def clientRecords(self):
        """ provides client recorders

        :returns: list of client recorders
        :rtype: :obj:`list` <:obj:`str`>
        """
        res = self.description
        dds = {}

        for cpg in res:
            for cp, dss in cpg.items():
                if isinstance(dss, dict):
                    if cp in self.cplist \
                       or cp in self.mcplist \
                       or cp in self.dslist \
                       or cp in self.acplist:
                        for ds, values in dss.items():
                            for vl in values:
                                if len(vl) > 1 and vl[1] == 'CLIENT':
                                    dds[ds] = vl[2]
        return list(set(dds.values()) - set(self.fullnames.values()) -
                    set(self.recorder_names))

    def stepComponents(self):
        """ provides components with step datasources

        :returns: list of components with step datasources
        :rtype: :obj:`list` <:obj:`str`>
        """
        res = self.description
        cpset = set()
        for cpg in res:
            for cp, dss in cpg.items():
                if isinstance(dss, dict):
                    for values in dss.values():
                        for vl in values:
                            if len(vl) > 0 and vl[0] == 'STEP':
                                cpset.add(cp)
                                break
                        else:
                            continue
                        break
        return list(cpset)

    def clientDataSources(self):
        """ provides client datasources

        :returns: list of client datasources
        :rtype: :obj:`list` <:obj:`str`>
        """
        res = self.dsdescription
        dsset = set()
        for jdsg in res:
            dsg = json.loads(jdsg)
            if isinstance(dsg, dict):
                if dsg['dstype'] == 'CLIENT':
                    dsset.add(dsg['dsname'])
        return list(dsset)

    #: (:obj:`list` <:obj:`str`>) provides disable datasources
    ddsdict = property(__disableDataSources,
                       doc='provides disable datasources')

    def __components(self):
        """ provides selected components

        :returns: list of selected components
        :rtype: :obj:`list` <:obj:`str`>
        """
        if isinstance(self.cpgroup, dict):
            return [cp for cp in self.cpgroup.keys() if self.cpgroup[cp]]
        else:
            return []

    #: (:obj:`list` <:obj:`str`>) provides selected components
    cplist = property(__components,
                      doc='provides selected components')

    def __datasources(self):
        """ provides selected datasources

        :returns: list of selected datasources
        :rtype: :obj:`list` <:obj:`str`>
        """
        if isinstance(self.dsgroup, dict):
            return [cp for cp in self.dsgroup.keys() if self.dsgroup[cp]]
        else:
            return []

    #: (:obj:`list` <:obj:`str`>) provides selected datasources
    dslist = property(__datasources,
                      doc='provides selected datasources')



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# Main window of the application

""" main window application dialog """

import os

qt_api = os.getenv("QT_API", os.getenv('DEFAULT_QT_API', 'pyqt5'))
if qt_api != 'pyqt4':
    try:
        __import__("PyQt5")
        qt_api = 'pyqt5'
    except Exception:
        __import__("PyQt4")
        qt_api = 'pyqt4'
else:
    __import__("PyQt4")
    qt_api = 'pyqt4'


__all__ = ["qt_api"]



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2014-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# Class with frames

""" main window application dialog """

import json

import logging
from .Element import CP, DS

#: (:obj:`logging.Logger`) logger object
logger = logging.getLogger(__name__)


class Frames(object):
    """ element group frames
    """

    def __init__(self, settings=None, ds=False, cp=False):
        """ constructor

        :param settings: frame settings
        :type settings: :obj:`list` < :obj:`list` < :obj:`list` \
                         [:obj:`str`, :obj:`int`] > > >
        :param ds: if datasource default frame
        :type ds: :obj:`bool`
        :param cp: if component default frame
        :type cp: :obj:`bool`
        """
        #: (:obj:`int`) datasource frame id
        self.dsid = None
        #: (:obj:`int`) component frame id
        self.cpid = None
        #: (:obj:`dict` <:obj:`int`, :obj:`str`>) (id, label) group dictionary
        self.__dct = None
        self.__settings = None
        #: ([[[:obj:`str`, :obj:`int`]]]) default datasource frame config
        self.defaultds = [[["Channels", 0]]]
        #: ([[[:obj:`str`, :obj:`int]`]]) default component frame config
        self.defaultcp = [[["Components", 1]]]
        self.set(settings, ds, cp)

    def set(self, settings, ds=False, cp=False):
        """ sets frame

        :param settings: frame settings
        :type settings: frame settings
        :param ds: if datasource frame
        :type ds: :obj:`bool`
        :param cp: if component frame
        :type cp: :obj:`bool`
        """
        if settings:
            mysettings = json.loads(settings)
        else:
            mysettings = None
        self.__dct = {}
        self.__settings = [self.defaultds, self.defaultcp]
        try:
            if mysettings is not None:
                groups = set()
                if ds or cp:
                    for frame in mysettings:
                        for column in frame:
                            for group in column:
                                groups.add(group[1])
                if cp and CP not in groups:
                    mysettings.insert(0, self.defaultcp)
                if ds and DS not in groups:
                    mysettings.insert(0, self.defaultds)
                self.__settings = list(mysettings)
            self.__makedict()

            ids = set(self.ids())
            if len(ids) < 1:
                self.__settings = [self.defaultds]
                self.__makedict()

        except Exception:
            self.__settings = [self.defaultds, self.defaultcp]
            self.__makedict()

        ids = list(set(self.ids()))
        self.dsid = ids[0]
        if len(ids) > 1:
            self.cpid = ids[1]
        else:
            self.cpid = ids[0]
        logger.debug("DSid = %s, CPid = %s " % (self.dsid, self.cpid))
        logger.debug(self.__dct)
        logger.debug(self.ids())

    def __makedict(self):
        """ create group dictionary
        """
        for frame in self.__settings:
            for column in frame:
                for group in column:
                    if group[1] in self.__dct.keys():
                        raise Exception("Duplicated Frame ID")
                    self.__dct[group[1]] = (group[0])

    def __iter__(self):
        """ provides frame iterator

        :returns: frame iterator
        :rtype: listiterator
        """
        return iter(self.__settings)

    def name(self, fid):
        """ provides frame label from id

        :param fid: frame id
        :type fid: :obj:`int`
        :returns: frame label
        :rtype: :obj:`str`
        """
        if fid in self.__dct.keys():
            return self.__dct[fid][0]

    def ids(self, name=None):
        """ provides frame id from label

        :parama name: frame label
        :type name: :obj:`str`
        :returns: frame id
        :rtype: :obj:`int`
        """
        if not name:
            return self.__dct.keys()
        res = set()
        for mid in self.__dct.keys():
            if name == self.__dct[mid]:
                res.add(mid)
        return res
